"""Machine learning processes"""
from pyppl import Proc, Box
from . import params, rimport
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pRegressTrain():
	"""
	@name:
		pRegressTrain
	@description:
		Train a regression model
	@input:
		`infile:file`: The input file (regression done on columns)
		`fmfile:file`: The file defining the regression formula.
			```
			Case1   lm   Y ~ X1 + X2
			Case2   glm  Y ~ X1 + X2 + X1:X2
			```
			- `Case` column can be omitted, if so, `Case1`, `Case2` will be added automatically.
			- No head.
		`casefile:file`: Defining cases:
			```
			R1   Case1
			R2   Case1
			... ...
			Rm   Case1
			Rm   Case2
			Rm+1 Case2
			... ...
			Rn   Case2
			```
			- One row can bed used for multiple cases
			- No head.
	@output:
		`outfile:file`: The output file with the summary of the models
		`outdir:dir`  : The output directory containing the summaries, models and plots
	@args:
		`plot`   : Plot the model? Default: `False`
		`inopts` : Input options for input file. Default: `Box(cnames = True, rnames = True)`
		`devpars`: The device parameters for the plot. Default: `Box(res = 300, height = 4000, width = 4000)`
	"""
	pRegressTrain        = Proc(desc = 'Train a regression model')
	pRegressTrain.input  = 'infile:file, fmfile:file, casefile:file'
	pRegressTrain.output = [
		'outfile:file:{{i.infile | fn2}}.regress/{{i.infile | fn2}}.regress.txt',
		'outdir:dir:{{i.infile | fn2}}.regress'
	]
	pRegressTrain.args.plot    = False
	pRegressTrain.args.cov     = ''
	pRegressTrain.args.yval    = 'numeric'
	pRegressTrain.args.save    = True
	pRegressTrain.args.glmfam  = 'binomial'
	pRegressTrain.args.inopts  = Box(rnames = True, cnames = True)
	pRegressTrain.args.devpars = Box(res = 300, height = 4000, width = 4000)
	pRegressTrain.envs.rimport = rimport
	pRegressTrain.lang         = params.Rscript.value
	pRegressTrain.script       = "file:scripts/mlearn/pRegressTrain.r"
	return pRegressTrain

@procfactory
def _pRegressPred():
	"""
	@name:
		pRegressPred
	@description:
		Use a trained linear regression model to predict
	@input:
		`infile:file`: The input file
		`model:file` : The trained model by `pRegressTrain`.
			- It can also be a directory containing models generated by `pRegressTrain`.
			- We will scan the model with file name: `*.<model>.rds`
			- Models have to be on the same column.
	@output:
		`outfile:file`: The output file, a copy of infile with Case, predicted value and items in `args.out` added.
		`outdir:dir`  : The output directory
	@args:
		`inopts` : The input options.
		`out`    : What to include in the output file
			- `prob`: Output the probabilities. Default: `True`
			- `auc` : Output AUCs. Default: `True` (only when Y column exists in infile)
			- If `auc` is True, then ROC is anyway plotted.
		`plot`   : Parameters to plot the ROC. Use `False` to disable. Default: `False`
			- Enabled only when Y column exists in infile
			- Could be a dict (Box) with keys:
			- `labels`:  show labels or not
			- `showAUC`: show AUC or not
			- `combine`: combine all roc in one figure?
		`ggs`    : The ggs items to be added for the plot. Default:
			```Box({
				'style_roc': {},
				# show legend at bottom right corner
				'theme#auc': {'legend.position': [1, 0], 'legend.justification': [1, 0]}
			})```
		`devpars`: The device parameters for the plot. Default: `Box(res = 300, height = 2000, width = 2000)`
		`inopts` : Options for reading the input file. Default: `Box(cnames = True, rnames = True, delimit = "\t")`
	"""
	pRegressPred           = Proc(desc = 'Use a trained linear regression model to predict')
	pRegressPred.input     = 'infile:file, model:file'
	pRegressPred.output    = [
		'outfile:file:{{i.infile | fn2}}.pred/{{i.infile | fn2}}.pred.txt',
		'outdir:dir:{{i.infile | fn2}}.pred'
	]
	pRegressPred.args.out  = Box(prob = True, auc = True)
	pRegressPred.args.plot = Box(labels = False, showAUC = True, combine = True)
	pRegressPred.args.ggs  = Box({
		'style_roc': {},
		# show legend at bottom right corner
		'theme#auc': {'legend.position': [1, 0], 'legend.justification': [1, 0]}
	})
	pRegressPred.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pRegressPred.args.inopts  = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t"
	)
	pRegressPred.envs.rimport = rimport
	pRegressPred.lang         = params.Rscript.value
	pRegressPred.script       = "file:scripts/mlearn/pRegressPred.r"
	return pRegressPred

@procfactory
def _pGlmTrain():
	"""
	@input:
		infile: The input file (Last column as Y)
	@output:
		outmodel: The output model (RData file)
		outdir  : The output directory containing model, plots and other files
	@args:
		plot   : Whether plot the glm probability. Default: `True`
		formula: The formula to perform the regression. Default: `None`.
			- If `None`, will use all first N-1 columns as features.
		inopts : The input options.
		yval   : The type of y values. Default: `categ`
			- `categ`  : categorical values
			- `prob`   : probabilities
			- `numeric`: numeric values
		devpars: The device parameters for the plot.
	"""
	return Box(
		desc   = 'Train a logistic regression model',
		lang   = params.Rscript.value,
		input  = 'infile:file',
		output = [
			'outmodel:file:{{i.infile | stem}}.glm/{{i.infile | stem}}.glm.RDS',
			'outdir:dir:{{i.infile | stem}}.glm'
		],
		args = Box(
			plot    = True,
			formula = None,
			devpars = Box(res = 300, height = 2000, width = 2000),
			params  = Box(family = 'binomial'),
			ggs     = Box(
				geom_smooth = Box({
					"method"     : "glm",
					"method.args": Box(family = "binomial"),
					"se"         : True
				})
			),
			inopts  = Box(
				cnames  = True,
				rnames  = True,
				delimit = "\t"
			),
			yval    = 'categ'))

@procfactory
def _pGlmTest():
	"""
	@input:
		`infile:file`: The input file
		`model:file` : The trained model by `pLogitRegTrain`
	@output:
		`outdir:dir`: The output directory
	@args:
		`inopts` : The input options.
		`outprob`: Also output probabilities? Default: True
	"""
	return Box(
		desc   = 'Test trained logistic regression model',
		lang   = params.Rscript.value,
		input  = 'infile:file, model:file',
		output = 'outdir:dir:{{i.infile | stem}}.glm.test',
		args   = Box(
			outprob = True,
			outauc  = True,
			params  = Box(),
			ggs     = Box(),
			devpars = Box(res = 300, height = 2000, width = 2000),
			inopts  = Box(
				cnames  = True,
				rnames  = True,
				delimit = "\t"
			)))

@procfactory
def _pRandomForestTrain():
	"""
	@input:
		`infile:file`: The input file (Last column as Y)
	@output:
		`outmodel:file`: The output model (RData file)
		`outdir:dir`   : The output directory containing model, plots and other files
	@args:
		`plot`   : Whether plot the feature importance. Default: `True`
		`formula`: The formula to perform the regression. Default: `None`.
			- If `None`, will use all first N-1 columns as features.
		`inopts` : The input options.
		`na`     : Replace NAs with? Default: `0`
		`devpars`: The device parameters for the plot.
	@requires:
		`r-randomForst`
	"""
	pRandomForestTrain = Proc(desc = "Train a random forest model.")
	pRandomForestTrain.input = 'infile:file'
	pRandomForestTrain.output = [
		'outmodel:file:{{i.infile | stem}}.rforest/{{i.infile | stem}}.rforest.rds',
		'outdir:dir:{{i.infile | stem}}.rforest'
	]
	pRandomForestTrain.args.plot    = True
	pRandomForestTrain.args.formula = None
	pRandomForestTrain.args.na      = 0
	pRandomForestTrain.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pRandomForestTrain.args.inopts  = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t"
	)
	pRandomForestTrain.envs.rimport = rimport
	pRandomForestTrain.lang         = params.Rscript.value
	pRandomForestTrain.script       = "file:scripts/mlearn/pRandomForestTrain.r"
	return pRandomForestTrain

@procfactory
def _pDecisionTreeTrain():
	"""
	@name:
		pDecisionTreeTrain
	@description:
		Train a decision tree model
	@input:
		`infile:file`: The input file (Last column as Y)
	@output:
		`outmodel:file`: The output model (RData file)
		`outdir:dir`   : The output directory containing model, plots and other files
	@args:
		`plot`   : Whether plot the feature importance. Default: `True`
		`formula`: The formula to perform the regression. Default: `None`.
			- If `None`, will use all first N-1 columns as features.
		`inopts` : The input options.
		`na`     : Replace NAs with? Default: `0`
		`devpars`: The device parameters for the plot. Default: `Box(res = 300, height = 2000, width = 2000)`
	@requires:
		`r-rpart`
	"""
	pDecisionTreeTrain = Proc(desc = "Train a decision tree model")
	pDecisionTreeTrain.input = 'infile:file'
	pDecisionTreeTrain.output = [
		'outmodel:file:{{i.infile | stem}}.dtree/{{i.infile | stem}}.dtree.rds',
		'outdir:dir:{{i.infile | stem}}.dtree'
	]
	pDecisionTreeTrain.args.plot    = True
	pDecisionTreeTrain.args.formula = None
	pDecisionTreeTrain.args.na      = 0
	pDecisionTreeTrain.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pDecisionTreeTrain.args.inopts  = Box(
		cnames  = True,
		rnames  = True,
		delimit = "\t"
	)
	pDecisionTreeTrain.envs.rimport = rimport
	pDecisionTreeTrain.lang         = params.Rscript.value
	pDecisionTreeTrain.script       = "file:scripts/mlearn/pDecisionTreeTrain.r"
	return pDecisionTreeTrain

@procfactory
def _pCrossValid():
	"""
	@input:
		`infile:file`: The input data file.
	@output:
		`outmodel:file`: The trained model in RDS format
		`outdir:dir`   : The output directory containing output model and plots.
	@args:
		`inopts` : The options to read the input file.
		`ctrl`   : Arguments for `trainControl`.
			- See https://topepo.github.io/caret/model-training-and-tuning.html#the-traincontrol-function
			- Available methods: "boot", "cv", "LOOCV", "LGOCV", "repeatedcv", "timeslice", "none" and "oob"
		`train`  : Arguments for `train` other than `data` and `trControl`.
			- See https://topepo.github.io/caret/model-training-and-tuning.html#basic-parameter-tuning
			- form: formula of the model,
			- method: rf, glm, etc
		`seed`   : The seed.
		`nthread`: # threads to use.
		`plots`  : Do types of plots.
			- `varimp` also available
			- You can also concatenate them using comma (`,`)
	@requires:
		`r-caret`
	"""
	return Box(
		desc   = 'Cross validation on the model',
		lang   = params.Rscript.value,
		input  = 'infile:file',
		output = [
			'outmodel:file:{{i.infile | fn2}}.{{args.train.method}}/{{i.infile | fn2}}.{{args.train.method}}.RDS',
			'outdir:dir:{{i.infile | fn2}}.{{args.train.method}}'
		],
		args = Box(inopts  = Box(cnames = True, rnames = True),
			ctrl    = Box(method = '', savePredictions = True, classProbs = True, verboseIter = True),
			train   = Box(form = None, method = '', metric = 'ROC'),
			seed    = None,
			nthread = 1,
			plots   = ['model', 'roc'], # varimp
			devpars = Box(res = 300, height = 2000, width = 2000)))




"""Utilities for bioprocs script"""
import re
import sys
from pathlib import Path
from tempfile import gettempdir
from contextlib import contextmanager
from colorama import Back
from pyparam import Helps, HelpAssembler
from pyppl import utils, Channel, PyPPL, Box
import bioprocs

def substrReplace(string, starts, lengths, replace):
	"""Replace substrings"""
	if not isinstance(starts, (tuple, list)):
		starts = [starts]
	if not isinstance(lengths, (tuple, list)):
		lengths = [lengths]
	assert len(starts) == len(lengths)
	if not isinstance(replace, (tuple, list)):
		replace = [replace] * len(starts)

	delta = 0
	for i, start in enumerate(starts):
		# adjust starts
		string = string[:start + delta] + replace[i] + string[start + lengths[i] + delta:]
		delta += len(replace[i]) - lengths[i]
	return string

def highlight(origin, query, incase = True):
	"""Highlight string with query string"""
	# get all occurrences of q
	if incase:
		occurs = [m.start() for m in re.finditer(query.lower(), origin.lower())]
	else:
		occurs = [m.start() for m in re.finditer(query, origin)]
	lengths = [len(query)] * len(occurs)
	return substrReplace(origin, occurs, lengths, [
		'{}{}{}'.format(Back.LIGHTRED_EX, origin[occur:occur+length], Back.RESET)
		for occur, length in zip(occurs, lengths)])

def highlightMulti(line, queries):
	"""Highlight a string with multiple queries"""
	for query in queries:
		line = highlight(line, query)
	return line

def subtractDict(bigger, smaller, prefix = ''):
	"""Subtract a dict from another"""
	ret = bigger.copy()
	for key, val in smaller.items():
		if key not in ret:
			continue
		if isinstance(ret[key], dict) and isinstance(val, dict) and ret[key] != val:
			ret[key] = subtractDict(ret[key], val, prefix + '  ')
		elif ret[key] == val:
			del ret[key]
	return ret

class Module:
	"""A module of bioprocs"""
	@staticmethod
	def modules():
		"""Get all modules"""
		return [module.stem
			for module in Path(bioprocs.__file__).parent.glob('*.py')
			if not module.stem.startswith('_')]

	@staticmethod
	def deindent(lines):
		"""Remove indent based on the first line"""
		indention = ''
		for line in lines:
			if not line:
				continue
			indention = line[:-len(line.lstrip())]
			break
		ret = []
		for line in lines:
			if not line:
				continue
			if not line.startswith(indention):
				raise ValueError('Unexpected indention at doc line:\n' + repr(line))
			ret.append(line[len(indention):].replace('\t', '  '))
		return ret

	def __init__(self, name):
		self.name   = name
		try:
			self.module = getattr(__import__('bioprocs', fromlist = [name]), name)
		except AttributeError as ex:
			if "has no attribute '%s'" % name in str(ex):
				raise AttributeError('No such module: %s' % name) from None
			else:
				raise
		self.desc   = self.module.__doc__ and self.module.__doc__.strip() or '[ Not documented. ]'
		self._procs = {}

	def procs(self):
		"""Get the processes of the module"""
		if self._procs:
			return self._procs
		for proc, factory in self.module._mkenvs.items():
			if len(proc) < 3 or proc[0] != '_' or proc[1] != 'p' \
				or not (proc[2].isdigit() or proc[2].isupper()):
				continue
			procobj = factory()
			self._procs[proc[1:]] = Process(procobj, self.module, factory.__doc__ and factory.__doc__.splitlines() or [])

		return self._procs

	def toHelps(self, helpsec):
		"""Send me to a Helps section"""
		helpsec.prefix = ''
		helpsec.add((self.name, '', self.desc))

	@contextmanager
	def loadProcs(self, index = None):
		"""Loading procs with indicators"""
		info = 'Collecting module%s: %s ...' % (index and (' ' + index) or '', self.name)
		print('\r' + info, end = '')
		yield self.procs()
		print(end = '\r')
		print(' ' * len(info), end = '')
		if index:
			idx, total = index.split('/')
			if idx == total:
				print(end = '\r')

	def toHelpsAsSec(self, helps):
		"""Add me to helps as a section"""
		helps.add(self.name + ': ' + self.desc, sectype = 'option', prefix = '')

	def toHelpsWithProcs(self, helps, nmods):
		"""Add me to helps with procs information"""
		self.toHelpsAsSec(helps)
		with self.loadProcs("%s/%s" % (len(helps), nmods)) as procs:
			for proc in procs.values():
				proc.toHelps(helps.select(self.name + ': '))

	def addToCompletions(self, comp):
		"""Add me to completions"""
		comp.addCommand(self.name, self.desc)
		with self.loadProcs() as procs:
			for pname, proc in procs.items():
				comp.addCommand(self.name + '.' + pname, proc.desc[0])
				proc.addToCompletions(comp.command(self.name + '.' + pname))

class Process:
	"""A bioprocs process"""
	def __init__(self, proc, module, doc):
		self.name    = proc.id
		self.module  = module
		self.proc    = proc
		self.desc    = self.proc.desc
		self.doc     = Module.deindent(doc) # list
		self._helps  = Helps()
		self._parsed = {}

	def toHelps(self, helpsec):
		"""Send me to a Helps section"""
		helpsec.prefix = ''
		if self.proc.origin == self.name:
			helpsec.add((self.name, '', self.desc))
		else:
			helpsec.add((self.name, '', 'Alias of: %s' % self.proc.origin))

	@staticmethod
	def _parseOptionSec(sec):
		ret      = {}
		sec      = Module.deindent(sec)
		lastdesc = []
		for line in sec:
			if not line:
				continue
			if line[0] in ('\t', ' '):
				lastdesc.append(line)
				continue
			parts = utils.split(line, ':', trim = False)
			nametype = parts.pop(0)
			lastdesc = [':'.join(parts).strip()]
			nametype = nametype.strip('`) ')
			if ' (' in nametype:
				name, typ = nametype.split(' (', 1)
			elif ':' in nametype:
				name, typ = nametype.split(':', 1)
			else:
				name, typ = nametype.strip(), ''
			ret[name.strip()] = (typ.strip(), lastdesc)
		return ret

	@staticmethod
	def _parsePlainSec(sec):
		return Module.deindent(sec)

	def parsed(self):
		"""Get parsed doc"""
		if self._parsed:
			return self._parsed
		if not self.doc:
			self._parsed = {}
			return self._parsed
		self._parsed = {}
		lastsec = []
		for line in self.doc:
			if line.startswith('@'):
				secname = line.strip('@: ')
				lastsec = []
				self._parsed[secname] = lastsec
			else:
				lastsec.append(line)
		for key, sec in self._parsed.items():
			if key in ('input', 'output', 'args'):
				self._parsed[key] = Process._parseOptionSec(sec)
			else:
				self._parsed[key] = Process._parsePlainSec(sec)
		return self._parsed

	@staticmethod
	def defaultVal(val, prefix = 'Default: ', indent = 0):
		"""Formatted value in help"""
		ret = []
		lines = utils.formatDict(val if val != '' else "''", 0).splitlines()
		for i, line in enumerate(lines):
			if i == 0:
				line = prefix + line
				if len(lines) > 1:
					line += ' \\'
				ret.append(line)
			else:
				ret.append(' ' * indent + line + ' \\')
		return ret

	def inputs(self):
		"""Get the input keys and types by definitions"""
		if isinstance(self.proc.config.input, dict):
			inkeys = self.proc.config.input.keys()
		elif isinstance(self.proc.config.input, str):
			inkeys = utils.split(self.proc.config.input, ',')
		else:
			inkeys = self.proc.config.input
		ret = utils.OBox()
		for inkey in inkeys:
			if ':' not in inkey:
				inkey += ':var'
			inname, intype = inkey.split(':', 1)
			ret[inname] = intype
		return ret

	def outputs(self):
		"""Get the input keys, types and default values by definitions"""
		if isinstance(self.proc.config.output, dict):
			outs = list(self.proc.config.output.keys())
		elif isinstance(self.proc.config.output, str):
			outs = utils.split(self.proc.config.output, ',')
		else:
			outs = self.proc.config.output
		ret = utils.OBox()
		for out in outs:
			parts = utils.split(out, ':')
			if len(parts) == 2:
				outname, outype, default = parts[0], 'var', parts[1]
			else:
				outname, outype, default = parts
			ret[outname] = (outype, default)
		return ret

	def addToCompletions(self, comp):
		"""Add me to completions"""
		for inname in self.inputs():
			docdesc = self.parsed().get('input', {}).get(inname, (None, ['Input %s' % inname]))[1]
			comp.addOption('-i.' + inname, docdesc[0])
		for outname, outypedeft in self.outputs().items():
			docdesc = self.parsed().get('output', {}).get(
				outname, ('var', Process.defaultVal(outypedeft[1])[0]))[1]
			comp.addOption('-o.' + outname, docdesc[0])
		for key in self.proc.args:
			docdesc = self.parsed().get('args', {}).get(key, ('auto', ['[ Not documented. ]']))[1]
			comp.addOption('-args.' + key, docdesc[0])
		# add -config.
		comp.addOption('-config.', 'Pipeline configrations, such as -config._log.file')

	def helps(self):
		"""Construct help page using doc"""
		if self._helps:
			return self._helps

		self._helps.add('Name',
			Path(self.module.__file__).stem + '.' + self.name + '%s [lang = %s]' % (
				'(%s)' % self.proc.origin if self.name != self.proc.origin else '',
				self.proc.lang))
		self._helps.add('Description', self.parsed().get('description') or self.desc)

		# input
		self._helps.add('Input options (Use \':list\' for multi-jobs)',
			sectype = 'option', prefix = '-')
		for inname, intype in self.inputs().items():
			doctype, docdesc = self.parsed().get('input', {}).get(inname, ('var', 'Input %s' % inname))
			intype = intype or doctype or 'var'
			self._helps.select('Input options').add(('-i.' + inname, '<%s>' % intype, docdesc))

		# output
		self._helps.add('Output options (\'exdir\' implied if path specified)',
			sectype = 'option', prefix = '-')

		for outname, outypedeft in self.outputs().items():
			outype, outdeft = outypedeft
			doctype, docdesc = self.parsed().get('output', {}).get(
				outname, ('var', Process.defaultVal(outdeft)))
			outype = outype or doctype or 'var'

			if not docdesc or ('default: ' not in docdesc[-1].lower() and len(docdesc[-1]) > 20):
				docdesc.extend(Process.defaultVal(outdeft, indent = 5))
			elif 'default: ' not in docdesc[-1].lower():
				defaults = Process.defaultVal(outdeft, indent = len(docdesc[-1]) + 6)
				docdesc[-1] += ' ' + defaults.pop(0)
				docdesc.extend(defaults)
			self._helps.select('Output options').add(('-o.' + outname, '<%s>' % outype, docdesc))

		# args
		self._helps.add('Process arguments', sectype = 'option', prefix = '-')
		for key, val in self.proc.args.items():
			doctype, docdesc = self.parsed().get('args', {}).get(key, (None, ['[ Not documented. ]']))
			doctype = doctype or type(val).__name__
			if doctype == 'NoneType':
				doctype = 'auto'
			if not docdesc or ('default: ' not in docdesc[-1].lower() and len(docdesc[-1]) > 20):
				docdesc.extend(Process.defaultVal(val, indent = 5))
			elif 'default: ' not in docdesc[-1].lower():
				defaults = Process.defaultVal(val, indent = len(docdesc[-1]) + 6)
				docdesc[-1] += ' ' + defaults.pop(0)
				docdesc.extend(defaults)

			self._helps.select('Process arguments').add((
				'-args.' + key,
				'<%s>' % doctype if str(doctype).lower() != 'bool' else '[bool]',
				docdesc))

		self._helps.add('Other options', sectype = 'option', prefix = '-')
		# process properties
		self._helps.select('Other options').add(
			('-<prop>', '', 'Process properties, such as -forks, -exdir, -cache ...'))

		# pipeline configurations
		self._helps.select('Other options').add(
			('-config.<subconf>[.subconf]', '',
			 'Pipeline configrations, such as -config._log.file'))

		# help
		self._helps.select('Other options').addParam(
			bioprocs.params[bioprocs.params._hopts[0]], bioprocs.params._hopts, ishelp = True)

		return self._helps

	@staticmethod
	def _updateArgs(args):
		# replace ',' with '.' in key
		# a bug of python-box, copy lost metadata
		#ret = args.copy() # try to keep the type
		ret = Box()
		for key, val in args.items():
			ret[key.replace(',', '.')] = Process._updateArgs(val) \
				if isinstance(val, dict) else val
		return ret

	def printHelps(self, error = None, halt = True):
		"""Print helps"""
		assembler = HelpAssembler()
		error = error or []
		if isinstance(error, str):
			error = [error]
		for err in error:
			print(assembler.error(err))
		print('\n'.join(assembler.assemble(self.helps())), end = '')
		if halt:
			sys.exit(1)

	def run(self, opts):
		"""Construct a pipeline with the process and run it"""
		if any(opts.get(h) for h in bioprocs.params._hopts) or \
			all(key in bioprocs.params._hopts for key in opts):
			self.printHelps()
		if not opts.get('i') or not isinstance(opts.i, dict):
			self.printHelps(error = 'No input specified.')

		indata = {}
		for inkey, intype in self.inputs().items():
			# We should allow some input options to be empty
			#if opts.i.get(inkey) is None:
			#	self.printHelps(error = 'Input "[-i.]%s" is not specified.' % inkey)

			# to be smart on "files"
			if intype in self.proc.IN_FILESTYPE and inkey in opts.i \
				and not isinstance(opts.i[inkey], list):
				indata[inkey + ':' + intype] = Channel.create([[opts.i[inkey]]])
			else:
				indata[inkey + ':' + intype] = Channel.create(opts.i.get(inkey))
		self.proc.input = indata

		if opts.get('o') is not None:
			if not isinstance(opts.o, dict):
				self.printHelps(error = 'Malformat output specification.')

			outdata = utils.OBox()
			for outkey, outypedeft in self.outputs().items():
				outype, outdeft = outypedeft
				if not opts.o.get(outkey):
					outdata[outkey + ':' + outype] = outdeft
					continue
				if outype not in ('file', 'dir'):
					outdata[outkey + ':' + outype] = opts.o[outkey]
					continue
				# try to extract exdir from output
				if '/' in opts.o[outkey]:
					out = Path(opts.o[outkey])
					if self.proc.exdir and out.parent != self.proc.exdir:
						raise ValueError('Cannot have output files/dirs with different parents as exdir.')
					self.proc.exdir = str(out.parent)
					outdata[outkey] = out.name
				else:
					outdata[outkey] = opts.o[outkey]
			self.proc.output = outdata

		if opts.get('args'):
			if not isinstance(opts.args, dict):
				self.printHelps(error = 'Malformat args specification.')
			self.proc.args.update(Process._updateArgs(opts.args))

		config = {
			'_log'  : {'file': None },
			'ppldir': Path(gettempdir()) / 'bioprocs.workdir'
		}
		config.update(Process._updateArgs(opts.get('config', {})))

		for key, val in opts.items():
			if key in bioprocs.params._hopts + ['i', 'o', 'args', 'config']:
				continue
			setattr(self.proc, key, val)

		PyPPL(config).start(self.proc).run()


class Pipeline:
	"""Assembled pipeline"""

	@staticmethod
	def pipelines():
		"""Get all available pipelines"""
		return [pplfile.stem[9:]
			for pplfile in (
				Path(bioprocs.__file__).parent / 'console' / 'pipeline').glob('bioprocs_*.py')
			if not pplfile.stem.startswith('_')]

	def __init__(self, name):
		self.name = name
		self.module = __import__('bioprocs.console.pipeline', fromlist = ['bioprocs_' + name])
		self.module = getattr(self.module, 'bioprocs_' + name)
		self.desc = self.module.__doc__ and self.module.__doc__.strip() or '[ Not documented. ]'

	def run(self):
		"""Run the pipeline"""
		prog = 'bioprocs ' + self.name
		sys.argv = [prog] + sys.argv[2:]
		bioprocs.params._prog = prog
		bioprocs.params._assembler.progname = prog

		self.module.main()

	def addToCompletions(self, comp):
		"""Add me to completions"""
		comp.addCommand(self.name, self.desc)
		self.module.params._addToCompletions(
			comp.command(self.name), withtype = False, alias = True)



from pyppl import Box
from bioprocs.utils import cmd, cmdargs, logger

infile  = {{i.infile | quote}}
sqlfile = {{i.sqlfile | quote}}
outfile = {{o.outfile | quote}}
argssql = {{args.sql | quote}}
inopts  = {{args.inopts | repr}}
outopts = {{args.outopts | repr}}

if sqlfile:
	with open(sqlfile) as f:
		sql = ' '.join(f.readlines()).strip()
	if argssql:
		logger.warning('`args.sql` is ignored, as `i.sqlfile` is provided.')
else:
	sql = argssql

if not sql:
	raise ValueError('One of `i.sqlfile` and `args.sql` is requied.')

params = {
	'H': inopts.cnames,
	'd': inopts.delimit,
	'e': inopts.encoding,
	'z': (inopts.gz == 'auto' and infile.endswith('.gz')) or inopts.gz is True,
	'D': outopts.delimit if outopts.delimit is not None else inopts.delimit,
	'O': outopts.cnames if outopts.cnames is not None else inopts.cnames,
	'E': outopts.encoding if outopts.encoding is not None else inopts.encoding
}
c = cmd.Cmd(['cat', infile]).pipe('q {} {} > {!r}'.format(
	cmdargs(params), 
	cmdargs({' ': sql})[2:],
	outfile
), shell = True).run()
logger.info('Done: %s', c.cmd)


from os import path, readlink
from bioprocs.utils import shell2 as shell, gztype

def check(ref):
	if not ref or not path.exists(ref):
		raise Exception('Reference file not exists: %s' % ref)

def checkIndex(refindex):
	if not refindex:
		refindex = [refindex]
	return all([path.exists(ri) for ri in refindex])

def buildIndex(ref, cmd, ref2 = None, cmd2 = None):
	try:
		runcmd(cmd)
		return ref
	except Exception:
		try:
			if not path.exists(ref2):
				shell.ln_s(ref, ref2)
			runcmd(cmd2)
			return ref2
		except Exception:
			return None

def bamIndex(bam, ext = '.bam.bai', samtools = 'samtools', nthread = 1):
	"""
	Index bam files
	If bam file is a link, try to find the index file in its orginal directory or
	its realpath directory
	If nothing found, try to create the index file using samtools
	@params:
		`ext`: The expected extension of index file. Default: `.bam.bai`
			- Some tools requird `XXX.bai` without `.bam`
		`samtools`: The path to samtools. Default: `samtools`
			- If it's None, then an exception will raised instead of creating the index file
		`nthread`: The # threads used to create the index file. Default: `1`
	"""
	if not ext.startswith('.'):
		ext = '.' + ext
	# /path/to/some.bam -> some.bam
	bname = path.basename(bam)
	# /path/to/some.bam -> /path/to/
	dname = path.dirname(bam)
	# some.bam -> some
	fname = path.splitext(bname)[0]
	# some -> some
	# [1]some -> some
	rname = fname.split(']', 1)[1] if fname.startswith('[') else fname

	shell.load_config(samtools = samtools)
	# /path/to/some.bam.bai
	expectedIndex = path.join(dname, fname + ext)
	if path.isfile(expectedIndex):
		return
	# if bam is not a link, there is nowhere else to find index, create it using samtools
	if not path.islink(bam):
		if samtools:
			shell.samtools.index(b = True, _out = expectedIndex, **{'@': nthread})
		else:
			raise ValueError('Index not found: {}'.format(bam))
		return
	# find the index in original directory
	origbam   = readlink(bam)
	origIndex = path.splitext(origbam)[0] + ext
	if path.isfile(origIndex):
		shell.ln_s(origIndex, expectedIndex)
		return
	# find the index in realpath directory
	realbam   = path.realpath(bam)
	realIndex = path.splitext(realbam)[0] + ext
	if path.isfile(realIndex):
		shell.ln_s(realIndex, expectedIndex)
		return
	# if all failed, create it
	if samtools:
		shell.samtools.index(b = True, _out = expectedIndex, **{'@': nthread})
	else:
		raise ValueError('Index not found: {}'.format(bam))

def tabixIndex(filename, type, tabix = 'tabix'):
	# /path/to/some.vcf -> some.vcf
	# /path/to/some.vcf.gz -> some.vcf
	bname = path.basename(filename[:-3]) if filename.endswith('.gz') else path.basename(filename)
	# /path/to/some.bam -> /path/to/
	dname = path.dirname(filename)
	# some.vcf -> some
	# some.vcf.gz -> some
	fname = path.splitext(bname[:-3] if bname.endswith('.gz') else bname)[0]
	# some -> some
	# [1]some -> some
	rname = fname.split(']', 1)[1] if fname.startswith('[') else fname

	expectedIndex = path.join(dname, fname + '.%s.gz.tbi' % type)
	if path.isfile(expectedIndex):
		return filename if filename.endswith('.gz') else filename + '.gz'
	
	shell.load_config(tabix = tabix)
	# type could bed3, bed6..
	ptype = 'bed' if type.startswith('bed') else type
	gt = gztype(filename)
	if gt == 'bgzip':
		if path.islink(filename):
			linkfile = readline(filename)
			if path.isfile(linkfile + '.tbi'):
				shell.ln_s(linkfile + '.tbi', expectedIndex)
				return filename
			readfile = path.realpath(filename)
			if path.isfile(realfile + '.tbi'):
				shell.ln_s(realfile + '.tbi', expectedIndex)
				return realfile
		shell.fg.tabix(p = ptype, _ = filename)
		return filename
	if gt == 'gzip':
		bgzfile = path.join(dname, bname + '.bgz.' + type)
		shell.gunzip(filename, c = True, _out = bgzfile)
		shell.bgzip(bgzfile)
		shell.fg.tabix(p = ptype, _ = bgzfile)
		return bgzfile + '.gz'
	shell.bgzip(filename, c = True, _out = filename + '.gz')
	shell.fg.tabix(p = ptype, _ = filename + '.gz')
	return filename + '.gz'

def vcfIndex(vcf, tabix = 'tabix'):
	return tabixIndex(vcf, 'vcf', tabix)

def bedIndex(bed, tabix = 'tabix'):
	return tabixIndex(bed, 'bed', tabix)



import vcf
from pyppl import Box
from copy import copy

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
helper  = {{args.helper | repr}}
if not isinstance(helper, list):
	helper = [helper]
helper  = [line for line in helper if line]
exec('\n'.join(helper), globals())

reader    = vcf.Reader(filename=infile)
readerops = {{args.reader}}
newreader = readerops(copy(reader)) if callable(readerops) else None
writer    = vcf.Writer(open(outfile, 'w'), newreader if newreader else reader)
recordops = {{args.record}}

for record in reader:
	formats = record.FORMAT.split(':')
	samples = [
		Box(data = Box(zip(formats, [getattr(sample.data, fm) for fm in formats]))) 
		for sample in record.samples
	]
	rec     = recordops(record, writer, samples)
	for i, sample in enumerate(record.samples):
		sample.data = sample.data._make(list(samples[i].data.values()))
	if rec:
		writer.write_record(rec)
writer.close()



from pyppl import Box
from os import path
from bioprocs.utils import shell, logger

infile     = {{i.manifile | quote}}
outdir     = {{o.outdir | quote}}
params     = {{args.params}}
nthread    = {{args.nthread | int}}
token      = {{args.token | repr}}
gdc_client = {{args.gdc_client | repr}}

gdc = shell.Shell({'gdc': gdc_client}, subcmd = True, equal = ' ').gdc

dftargs = Box({
	'retry-amount': '3',
	'debug': False,
	'log-file': path.join(outdir, 'gdc-client.log')
})
dftargs.update(params)
dftargs.m = infile
dftargs.n = nthread
dftargs.d = outdir
dftargs._ = ['shraw:1>&2']
if token:
	dftargs.t = token
gdc.download(**dftargs).run()

# check if all the data sucessfully downloaded
logger.warning('Checking if all files in manifest has been sucessfully downloaded ...')
with open(infile) as fin:
	ids = [line.split()[0] for line in fin if line.strip() and not line.startswith('id')]
logger.warning('Got %s file ids', len(ids))
logger.warning('%s files failed', sum(int(not path.isdir(path.join(outdir, i))) for i in ids))

del dftargs['m']
for i in ids:
	if not path.isdir(path.join(outdir, i)):
		logger.warning('FAILED: {}'.format(i))
		logger.warning('- Trying to find it in https directory ...')
		destfile = path.join(outdir, 'https:', 'api.gdc.cancer.gov', 'data', i)
		if path.exists(destfile):
			logger.warning('- Found, save it.')
			shell.mv(destfile, path.join(outdir, i))
		else:
			logger.warning('- Not found, try to download it.')
			dftargs['_'] = [i, 'shraw:1>&2']
			gdc.download(**dftargs).run()
			shell.mv(destfile, path.join(outdir, i))

if path.isdir(path.join(outdir, 'https:')):
	shell.rmrf(path.join(outdir, 'https:'))



import sys
from modkit import Modkit

import cmdy

DEFAULT_CONFIG = dict(
	default = dict(_raise = True),

	bedtools  = dict(_prefix = '-'),
	biobambam = dict(_sep = '=', _prefix = ''),
	bowtie2   = dict(_dupkey = True),
	dtoxog    = dict(_out = cmdy.DEVERR, _prefix = '-'),
	sort      = dict(_sep = '', _dupkey = True),
	gatk      = dict(_dupkey = True),
	liftover  = dict(_prefix = '-', _sep = '='),
	oncotator = dict(_sep = 'auto'),
	maf2vcf   = dict(_sep = ' '),

	# As of picard 2.20.5-SNAPSHOT
	# it's changing in the futher. See: https://github.com/broadinstitute/picard/wiki/Command-Line-Syntax-Transition-For-Users-(Pre-Transition)
	# Future one should be:
	# picard = dict(_sep = ' ', _prefix = '-')
	picard    = dict(_sep = '=', _prefix = ''),
	plink     = dict(_out = cmdy.DEVERR),
	pyclone   = dict(_raw = True),
	samtools  = dict(_prefix = '-'),
	snpeff    = dict(_prefix = '-'),
	vcfanno   = dict(_prefix = '-'),
)
cmdy.config._load(DEFAULT_CONFIG)

def _modkit_delegate(name):
	return getattr(cmdy, name)

# run command at foreground
fg   = cmdy(_fg = True, _debug = True)
bg   = cmdy(_bg = True, _debug = True)
out  = cmdy(_out = '>')
pipe = cmdy(_pipe = True)

## aliases
rm_rf  = cmdy.rm.bake(r = True, f = True)
ln_s   = cmdy.ln.bake(s = True)
kill_9 = cmdy.kill.bake(s = 9)
wc_l   = cmdy.wc.bake(l = True)

runcmd = lambda cmd: cmdy.bash(c = cmd)

def load_config(conf = None, **kwargs):
	conf = conf or {}
	conf.update(kwargs)
	conf2load = {'default': DEFAULT_CONFIG['default']}
	for key, val in conf.items():
		conf2load[key] = DEFAULT_CONFIG.get(key, {}).copy()
		conf2load[key].update(val if isinstance(val, dict) else {'_exe': val})

	cmdy.config._load(conf2load)
	fg.config._load(conf2load)
	out.config._load(conf2load)

Modkit()



from os import path, makedirs
from glob import glob

{{runcmd}}
{{params2CmdArgs}}

# generate random regions
btparams = {}

btparams['l']    = {{args.len}}
btparams['n']    = 2*{{args.nperm}} # some regions do not have data
btparams['seed'] = {{i.seed}}
btparams['g']    = {{args.gsize | quote}}

randregfileUnsorted = path.join({{job.outdir | quote}}, 'randregions.bed.unsorted')
randregfile         = path.join({{job.outdir | quote}}, 'randregions.bed')
randregdir          = path.join({{job.outdir | quote}}, 'randregions')
cmd = '{{args.bedtools}} random %s > "%s"' % (params2CmdArgs(btparams, dash='-', equal=' '), randregfileUnsorted)
runcmd(cmd)

cmd = 'sort -k1,1 -k2,2n "%s" > "%s"' % (randregfileUnsorted, randregfile)
runcmd(cmd)

if not path.isdir(randregdir): makedirs(randregdir)
chroms = []
with open(randregfile) as f:
	prevchr = ''
	handler = None
	for line in f:
		if not line.strip(): continue
		chrom = line.split('\t')[0]
		if chrom != prevchr:
			chroms.append(chrom)
			if handler: handler.close()
			handler = open(path.join(randregdir, chrom + '.bed'), 'w')
			handler.write(line)
			prevchr = chrom
		else:
			handler.write(line)
	if handler:
		handler.close()


bwparams = {}
bwparams['skip-median'] = True

outfileUnsorted   = "{{o.outfile}}.unsorted"
outfileUnfiltered = "{{o.outfile}}.unfiltered"
for i, chrom in enumerate(chroms):
	regfile = path.join(randregdir, chrom + '.bed')
	bwfile  = glob(path.join({{args.consvdir | quote}}, chrom + '.*'))[0]
	cmd = '{{args.bwtool}} summary "%s" "%s" /dev/stdout %s | cat >> "%s"' % (regfile, bwfile, params2CmdArgs(bwparams, dash='-', equal=' '), outfileUnsorted)
	runcmd(cmd)

cmd = 'sort -k8nr "%s" > "%s"' % (outfileUnsorted, outfileUnfiltered)
runcmd(cmd)

# remove regions without data
with open(outfileUnfiltered) as f, open({{out.outfile | quote}}, 'w') as fout:
	for line in f:
		if '\tNA\tNA\tNA' in line: continue
		fout.write(line)




from pyppl import Box
from bioprocs.utils import shell

infile   = {{ i.infile | quote}}
outfile  = {{ o.outfile | quote}}
n        = {{ args.n | repr}}
gsize    = {{ args.gsize | quote}}
seed     = {{ args.seed | repr}}
params   = {{ args.params | repr}}
bedtools = {{ args.bedtools | quote}}
bedtools = shell.Shell({'bedtools': bedtools}, subcmd = True, dash = '-', equal = ' ').bedtools

if n and n<1:
	total = shell.wcl(infile)
	n = int(total * n)

params.i = infile
params.g = gsize
params.seed = seed

if n:
	shuffle = bedtools.shuffle(**params)
	shuffle.pipe().head(n = n, _stdout = outfile)
else:
	params._stdout = outfile
	shuffle = bedtools.shuffle(**params)
	shuffle.run()



import sys
from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.reference import vcfIndex

infile   = {{i.infile | quote}}
samfile  = {{i.samfile | quote}}
outfile  = {{o.outfile | quote}}
bcftools = {{args.bcftools | quote}}
params   = {{args.params | repr}}
gz       = {{args.gz | repr}}
nthread  = {{args.nthread | repr}}
tabix    = {{args.tabix | quote}}

vcfIndex(infile, tabix = tabix)

shell.load_config(bcftools = bcftools)

params._    = infile
params.o    = outfile
params.O    = 'z' if gz else 'v'
if samfile:
	if path.isfile(samfile):
		params.S = samfile
	else:
		params.s = samfile
cmd = shell.bcftools.view(**params).cmd
sys.stderr.write("\n%s RUNNING %s\n%s\n%s\n\n" % ("-" * 40, "-" * 40, cmd, "-" *  89))



import re
from os import makedirs, path
from shutil import rmtree
from sys import stderr
from pyppl import Box
from bioprocs.utils import cmdargs, runcmd, mem2

# detemine default read group
rg = {{ args.rg }}
rg = {key.upper():val for key, val in rg.items()}
if not rg['ID']:
	g = re.search (r'[^a-zA-Z0-9]+(L\\d+)[^a-zA-Z0-9]+', "{{o.outfile | fn}}")
	rg['ID'] = g.group(1) if g else "{{o.outfile | fn}}.L{{job.index}}"
if not rg['SM']:
	rg['SM'] = "{{o.outfile | fn}}"

tmpdir    = path.join ("{{args.tmpdir}}", "{{proc.id}}.{{i.infile | fn}}.{{job.index}}")
if not path.exists (tmpdir):
	makedirs (tmpdir)

params = {{args.params}}
try:
{% case args.tool %}
	############## picard
	{% when 'picard' %}
	mem = mem2({{ args.mem | quote }})
	params['-Djava.io.tmpdir'] = tmpdir
	params['TMP_DIR'] = tmpdir
	params['I'] = {{i.infile | quote}}
	params['O'] = {{o.outfile | quote}}
	for k,v in rg.items():
		params['RG' + k] = v

	runcmd ('{{args.picard}} AddOrReplaceReadGroups %s %s' % (mem, cmdargs(params, dash='', equal='=')))

	############## bamutil
	{% when 'bamutil' %}
	params['RG'] = "@RG\\tID:%s\\t%s" % (rg['ID'], "\\t".join([k + ":" + v for k,v in rg.items() if k!='ID']))
	params['in'] = {{i.infile | quote}}
	params['out'] = {{o.outfile | quote}}

	runcmd ('{{args.bamutil}} polishBam %s' % cmdargs(params, equal = ' '))

{% endcase %}
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	rmtree (tmpdir)



import random
import vcf

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
keep    = {{args.keep | quote}}
upart   = {{args.upart | repr}}
gz      = {{args.gz | repr}}
if gz:
	outfile = outfile[:-3]

keeps = [k.strip() for k in keep.split(',')]

reader = vcf.Reader(filename=infile)
writer = vcf.Writer(open(outfile, 'w'), reader)

def writeRecord(records, ks):
	if len(records) == 1:
		writer.write_record(records[0])
	else:
		if 'bisnp' in ks:
			writeRecord([r for r in records if r.is_snp and len(r.ALT) == 1], [k for k in ks if k!='bisnp'])
		elif 'snp' in ks:
			writeRecord([r for r in records if r.is_snp], [k for k in ks if k!='snp'])
		elif 'bialt' in ks:
			writeRecord([r for r in records if len(r.ALT) == 1], [k for k in ks if k!='bialt'])
		elif 'last' in ks:
			writer.write_record(records[-1])
		elif 'random' in ks:
			writer.write_record(random.choice(records))
		else:
			writer.write_record(records[0])

buffer_name = None
buffer_recs = []
while True:
	try:
		record = reader.next()
		rid = '/'.join([str(getattr(record, i)) for i in upart])
		if buffer_name is None:
			buffer_name = rid
			buffer_recs.append(record)
		elif buffer_name == rid:
			buffer_recs.append(record)
		else:
			writeRecord(buffer_recs, keeps)
			buffer_name = rid
			buffer_recs = [record]
	except StopIteration:
		break
	except ValueError:
		continue

writeRecord(buffer_recs, keeps)
writer.close()

if gz:
	runcmd(['bgzip', outfile])





import csv
from os import path
from openpyxl import Workbook
from bioprocs.utils import log2pyppl

infiles  = {{i.infiles | repr}}
outfile  = {{o.outfile | quote}}
fn2sheet = {{args.fn2sheet}}

def tsv2sheet(wb, tsvfile):
	fn = path.splitext(path.basename(tsvfile))[0]
	ws = wb.create_sheet(fn if not callable(fn2sheet) else fn2sheet(fn))
	log2pyppl('Reading ' + tsvfile + ' ...')
	with open(tsvfile) as f:
		reader = csv.reader(f, delimiter = "\t")
		for row in reader:
			ws.append(row)

wb = Workbook()
wb.remove(wb.active) # remove default sheet
for infile in infiles:
	tsv2sheet(wb, infile)

wb.save(outfile)


from pyppl import Box
from deconstructSigs import DeconstructSigs, Settings

infile = {{i.infile | quote}}
outdir = {{o.outdir | quote}}
args   = {{args | repr}}

Settings.font_family = args.get('font_family', 'Arial')
Settings.font_weight = args.get('font_weight', 'bold')
Settings.sig_cutoff  = args.get('sig_cutoff' , .05)
Settings.err_thres   = args.get('err_thres'  , 1e-3)
ref = args.get('ref')

ds      = DeconstructSigs(maf = infile, hg19_fasta_path = ref, output_folder = outdir)
weights = ds.which_signatures()
ds.figures(weights, explanations=True)


from os import path
from pyppl import Box
from bioprocs.utils import shell

seed    = {{ i.seed | repr}}
outdir  = {{ o.outdir | quote}}
out     = path.join(outdir, "simsnps." + str(seed if isinstance(seed, int) else 'noseed'))
plink   = {{ args.plink | repr}}
ncases  = {{ args.ncases | repr}}
nctrls  = {{ args.nctrls | repr}}
nsnps   = {{ args.nsnps | repr}}
label   = {{ args.label | repr}}
dprev   = {{ args.dprev | repr}}
minfreq = {{ args.minfreq | repr}}
maxfreq = {{ args.maxfreq | repr}}
hetodds = {{ args.hetodds | repr}}
homodds = {{ args.homodds | repr}}
params  = {{ args.params | repr}}

plink = shell.Shell(dict(plink = plink), equal = ' ').plink

# write the parameter file
parfile = path.join(outdir, 'simsnps.par.txt')
with open(parfile, 'w') as f:
	f.write("{nsnps}\t{label}\t{minfreq}\t{maxfreq}\t{hetodds}\t{homodds}\n".format(
		nsnps = nsnps, label = label, minfreq = minfreq,
		maxfreq = maxfreq, hetodds = hetodds, homodds = homodds
	))
if isinstance(seed, int):
	params.seed = seed
params.simulate = parfile
params.out      = out

params['simulate-ncases']     = ncases
params['simulate-ncontrols']  = nctrls
params['simulate-prevalence'] = dprev

plink(**params).run()


from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts | repr}}
cfilter = {{args.filter}}

reader = TsvReader(infile, **inopts)
cnames = reader.cnames
if callable(cfilter):
	cnames = cfilter(cnames)
with open(outfile, 'w') as f:
	f.write("".join([cname + "\n" for cname in cnames if cname]))



from os import path, chdir
from glob import glob
from pyppl import Box
from bioprocs.utils import parallel, logger, shell2 as shell
from bioprocs.utils.tsvio2 import TsvWriter, TsvReader
from bioprocs.utils.reference import vcfIndex

infile       = {{i.infile | quote}}
outfile      = {{o.outfile | quote}}
tool         = {{args.tool | quote}}
vcf2maf      = {{args.vcf2maf | quote}}
vep          = {{args.vep | quote}}
tabix        = {{args.tabix | quote}}
vepDb        = {{args.vepDb | quote}}
filtervcf    = {{args.filtervcf | quote}}
ref          = {{args.ref | quote}}
bcftools     = {{args.bcftools | quote}}
genome       = {{args.genome | quote}}
tumoridx     = {{args.tumor | repr}}
withchr      = {{args.withchr | repr}}
nthread      = {{args.nthread | repr}}
params       = {{args.params | repr}}
oncotator    = {{args.oncotator | quote}}
oncotator_db = {{args.oncotator_db | quote}}

shell.load_config(vcf2maf = vcf2maf, bcftools = bcftools, oncotator = oncotator)

def extract_sample_from_vcf(vcf, sample, outvcf, nomiss = True):
	shell.fg.bcftools.view(s = sample, o = outvcf + '.tmp', _ = vcf)
	# exclude sites without genotypes
	# because this is split from a merged vcf file
	if nomiss:
		shell.fg.bcftools.view(_ = outvcf + '.tmp', g = '^miss', o = outvcf)
	else:
		shell.fg.bcftools.view(_ = outvcf + '.tmp', o = outvcf)
	shell.rm_rf(outvcf + '.tmp')

def run_vcf2maf_one(vcf, maf, tumor, normal = None, forks = nthread):
	params_one = params.copy()
	params_one['input-vcf']  = vcf
	params_one['output-maf'] = maf
	params_one['vep-data']   = vepDb
	params_one['vep-forks']  = forks
	params_one['filter-vcf'] = filtervcf
	params_one['ref-fasta']  = ref
	params_one['vep-path']   = path.dirname(shell.which(vep).strip())
	params_one['tumor-id']   = tumor
	params_one['normal-id']  = normal
	shell.fg.vcf2maf(**params_one)

def run_oncotator_one(vcf, maf, tumor, normal = None, forks = nthread):
	vcf = vcfIndex(vcf, tabix = tabix)
	vcf0 = vcf
	if normal: # split tumor and normal
		vcf_tumor  = vcf + '.' + tumor
		extract_sample_from_vcf(vcf, tumor, vcf_tumor, nomiss = False)
		vcf = vcf_tumor

	openblas_threads = max(nthread - forks, 1)
	params_one               = params.copy()
	params_one._             = [vcf, maf, 'hg19']
	params_one.v             = True
	params_one.input_format  = 'VCF'
	params_one.output_format = 'TCGAMAF'
	params_one.log_name      = maf + '.oncotator.log'
	# db_dir
	params_one['db-dir'] = oncotator_db
	# c
	txfiles = glob(path.join(oncotator_db, 'tx_exact_uniprot_matches*'))
	if txfiles:
		params.c = txfiles[0]

	# configs
	outdir = path.dirname(outfile)
	chdir(outdir)
	confs = glob(path.join(oncotator_db, '*.config'))
	for conf in confs:
		shell.ln_s(conf, path.join(outdir, path.basename(conf)))

	# don't use **params, otherwise input_format will be turned into input-format
	logger.info('See %s for oncotator logs.', params_one.log_name)
	shell.fg.oncotator(params_one, _env =  dict(
		OPENBLAS_NUM_THREADS = str(openblas_threads),
		OMP_NUM_THREADS      = str(openblas_threads),
		NUMEXPR_NUM_THREADS  = str(openblas_threads),
		MKL_NUM_THREADS      = str(openblas_threads)
	))

	# oncotator cannot put Matched_Norm_Sample_Barcode and allele information
	# add them if normal is specified
	if normal:
		# I suppose there isn't many somatic mutations, so I read the normal information first
		# into memory
		from pysam import VariantFile
		normal_infos = {}
		vfile        = VariantFile(vcf0)
		tum_index    = list(vfile.header.samples).index(tumor)
		norm_index   = list(vfile.header.samples).index(normal)
		for record in vfile:
			tum_alt_index1  = record.samples[tum_index]['GT'][0]
			tum_alt_index2  = record.samples[tum_index]['GT'][-1]
			tum_alt1        = record.alleles[tum_alt_index1]
			tum_alt2        = record.alleles[tum_alt_index2]
			chrom = record.chrom[3:] if record.chrom.startswith('chr') else record.chrom

			key = '{3}_{0.pos}_{0.ref}_{1}_{2}'.format(
				record, tum_alt1, tum_alt2, chrom)
			try:
				norm_alt_index1 = record.samples[norm_index]['GT'][0]
				norm_alt_index2 = record.samples[norm_index]['GT'][-1]
				normal_infos[key] = (record.alleles[norm_alt_index1], record.alleles[norm_alt_index2])
			except (TypeError, IndexError):
				pass

	reader = TsvReader(maf, cnames = True)
	writer = TsvWriter(maf + '.tmp')
	writer.cnames = reader.cnames

	# correct Start_position to Start_Position, End_position to End_Position
	#               ^                               ^
	writer.writeHead(lambda cnames: [
		'Start_Position' if cname == 'Start_position' else
		'End_Position' if cname == 'End_position' else cname
		for cname in cnames])
	for r in reader:
		key = '{r.Chromosome}_{r.Start_position}_{r.Reference_Allele}_{r.Tumor_Seq_Allele1}_{r.Tumor_Seq_Allele2}'.format(r = r)
		r.NCBI_Build = genome if r.NCBI_Build == '__UNKNOWN__' else r.NCBI_Build
		r.Tumor_Sample_Barcode = tumor
		r.Matched_Norm_Sample_Barcode = normal or 'NORMAL'
		if normal:
			try:
				r.Match_Norm_Seq_Allele1 = normal_infos[key][0]
				r.Match_Norm_Seq_Allele2 = normal_infos[key][1]
			except KeyError:
				r.Match_Norm_Seq_Allele1 = r.Reference_Allele
				r.Match_Norm_Seq_Allele2 = r.Reference_Allele
		if r.Variant_Type == 'INS' and r.Match_Norm_Seq_Allele1 == '-' and r.Tumor_Seq_Allele2:
			r.Match_Norm_Seq_Allele1 = r.Tumor_Seq_Allele2[0]
		if r.Variant_Type == 'INS' and r.Match_Norm_Seq_Allele2 == '-' and r.Tumor_Seq_Allele2:
			r.Match_Norm_Seq_Allele2 = r.Tumor_Seq_Allele2[0]
		if withchr and r.Chromosome[:3] != 'chr':
			r.Chromosome = 'chr' + r.Chromosome
		writer.write(r)

	shell.mv(maf + '.tmp', maf)

ones = dict(
	vcf2maf   = run_vcf2maf_one,
	oncotator = run_oncotator_one
)

def run(tool):
	global tumoridx

	one = ones[tool]

	vcfsams  = shell.bcftools.query(l = infile).splitlines()
	vcfsams  = [s for s in vcfsams if s.strip()]

	if len(vcfsams) == 1:
		one(infile, outfile, vcfsams[0])
	elif len(vcfsams) >= 2 and tumoridx in (0, 1, 'auto'):
		if len(vcfsams) == 2 and tumoridx == 'auto':
			from difflib import SequenceMatcher
			similarity = lambda a, b: SequenceMatcher(None, a, b).ratio()
			sim0       = similarity(path.basename(infile), vcfsams[0])
			sim1       = similarity(path.basename(infile), vcfsams[1])
			tumoridx   = int(sim1 > sim0)
			logger.info('Using sample %s(%s) as tumor', tumoridx + 1, vcfsams[tumoridx])

		tumor  = vcfsams[tumoridx]
		normal = vcfsams[1-tumoridx]
		one(infile, outfile, tumor, normal)
	else:
		# split vcf file
		splitdir = path.join(path.dirname(outfile), "splits")
		shell.mkdir(p = splitdir)
		mafdir = path.join(path.dirname(outfile), "mafs")
		shell.mkdir(p = mafdir)

		para = parallel.Parallel(nthread)
		para.run(extract_sample_from_vcf, [
			(infile, s, path.join(splitdir, "split{}.vcf".format(i+1)))
			for i, s in enumerate(vcfsams)
		])
		restThreads = int(float(nthread)/float(len(vcfsams)))
		restThreads = max(restThreads, 1)
		para.run(one, [
			(path.join(splitdir, "split" + str(i+1) + ".vcf"),
			 path.join(mafdir, "split{}.maf".format(i+1)),
			 s, None, restThreads)
			for i, s in enumerate(vcfsams)
		])
		del para

		# merge mafs
		shell.out.head(n = 1, _ = path.join(mafdir, "split1.maf")) > outfile
		for i, s in enumerate(vcfsams):
			shell.out.tail(n = '+2', _ = path.join(mafdir, "split{}.maf".format(i+1))) >> outfile

try:
	run(tool)
except KeyError:
	raise ValueError('Tool {!r} not supported.'.format(tool))



string = {{i.instr | quote}}

{% for b in args.breakOn %}
string = string.replace({{b | quote}}, '\n')
{% endfor %}

{% if args.trimLine %}
string = '\n'.join([x.strip() for x in string.splitlines()]) + '\n'
{% endif %}

with open('{{o.outfile}}', 'w') as fout:
	fout.write(string)


from pyppl import Box
from bioprocs.utils import mem2, logger, shell2 as shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
umfile   = {{o.umfile | quote}}
tool     = {{args.tool | quote}}
picard   = {{args.picard | quote}}
chain    = {{args.lochain | quote}}
ref      = {{args.ref | quote}}
params   = {{args.params | repr}}
mem      = {{args.mem | quote}}
tmpdir   = {{args.tmpdir | quote}}
bcftools = {{args.bcftools |quote}}

if not chain:
	logger.error('Chain file (args.lochain) not provided!')
	exit(1)

shell.load_config(picard = picard, bcftools = bcftools)

# picard LiftoverVcf -Xmx4g -Xms1g  I=TCGA-05-4382-10.vcf O=1.vcf CHAIN=liftovers/hg38ToHg19.over.chain.gz R=ucsc_hg19.fa REJECT=r.vcf

if tool == 'picard':

	params.I      = infile
	params.O      = outfile
	params.CHAIN  = chain
	params.REJECT = umfile
	params.R      = ref

	javamem = mem2(mem, 'java')
	for jm in javamem.split():
		params['-' + jm[1:]] = True

	params['-Djava.io.tmpdir'] = tmpdir
	shell.fg.picard.LiftoverVcf(**params)

	# check if samples being swapped 
	# picard LiftoverVcf did that and no options to correct
	samples_old = shell.bcftools.query(l = infile).strip().splitlines()
	samples_new = shell.bcftools.query(l = outfile).strip().splitlines()
	if samples_old == samples_new:
		exit(0)
	
	orders = [samples_new.index(sample) for sample in samples_old]
	tmpfile = outfile + '.sampleorder_incorrect'
	shell.mv(outfile, tmpfile)
	with open(tmpfile, 'r') as fin, open(outfile, 'w') as fout:
		for line in fin:
			if line.startswith('##'):
				fout.write(line)
				continue
			if line.startswith('#'):
				parts = line.split('\t')
				parts = parts[:9] + samples_old
				fout.write('\t'.join(parts) + '\n')
				continue
			parts = line.strip().split('\t')
			samdata = parts[9:]
			parts = parts[:9] + [samdata[i] for i in orders]
			fout.write('\t'.join(parts) + '\n')



import json
from pyppl import Box
from pyppl.utils import alwaysList
from medoo import Medoo, Raw, Field, Dialect
try:
	string_types = (str, unicode)
except NameError:
	string_types = (str, )

class DialectCache(Dialect):

	OPERATOR_MAP = {'~~': 'ilike'}

	@classmethod
	def ilike(klass, field, value):
		field = Raw('UPPER({})'.format(field))
		if not isinstance(value, (tuple, list)):
			value = [value]

		value = [Raw('UPPER({})'.format(klass.value(v))) for v in value]
		return klass.like(field, value, addperct = False)

	@classmethod
	def concat(klass, field, value):
		return '{} = {} || {}'.format(field, field, klass.value(value))


class Cache(object):

	ARRAY_DELIMIT = '|'

	DUMMY = dict(
		plain = dict(
			# compose the where conditions from the database column and query data
			# e.g. ('id', [1,2,3,4]) ==medoo==> 'id' IN (1,2,3,4)
			query  = lambda col, data: (col, Cache._uniqueData(data)),
			# tell whether a query item is found from the query results
			find   = lambda col, qitem, results: Cache._plainFinder(col, qitem, results),
			# how to save the data at the column
			insert = lambda col, data: (col, Cache._uniqueData(data)),
			update = lambda col, data, orig: (col, Cache._uniqueData(data)),
		),
		iplain = dict(
			# UPPER(name) in ('ABC', 'DEF', 'GHI')
			query  = lambda col, data:(col + '[~~]', Cache._expandLike(data)),
			# name.upper() in
			find   = lambda col, qitem, results: Cache._plainFinder(col, qitem, results, case = True),
			insert = lambda col, data: (col, Cache._uniqueData(data)),
			update = lambda col, data, orig: (col, Cache._uniqueData(data)),
		),
		array = dict(
			query  = lambda col, data: (col + '[~]', Cache._expandLike(data)),
			find   = lambda col, qitem, results: Cache._arrayFinder(col, qitem, results),
			insert = lambda col, data: (col, Cache._arrayJoiner(Cache._uniqueData(data, True))),
			update = lambda col, data, orig: (col + '[concat]', Cache._arrayJoiner(
				set(Cache._uniqueData(data, True)) - set(orig or [])
			)),
			result = lambda data: data if isinstance(data, list) or data is None else list(filter(None, data.split(Cache.ARRAY_DELIMIT)))
		),
		iarray = dict(
			query = lambda col, data: (col + '[~~]', Cache._expandLike(data)),
			find  = lambda col, qitem, results: Cache._arrayFinder(col, qitem, results, case = True),
			insert = lambda col, data: (col, Cache._arrayJoiner(Cache._uniqueData(data, True))),
			update = lambda col, data, orig: (col + '[concat]', Cache._arrayJoiner(
				set(Cache._uniqueData(data, True)) - set(orig or [])
			)),
			result = lambda data: data if isinstance(data, list) or data is None else list(filter(None, data.split(Cache.ARRAY_DELIMIT)))
		),
		json = dict(
			query  = lambda col, data: (col, [json.dumps(d, sort_keys = True) for d in set(data)]),
			find   = lambda col, qitem, results: Cache._jsonFinder(col, qitem, results),
			insert = lambda col, data: (col, json.dumps(data, sort_keys = True)),
			update = lambda col, data, orig: (col, json.dumps(data, sort_keys = True)),
			result = lambda data: data if not isinstance(data, string_types) else json.loads(data)
		)
	)

	def __init__(self, dbfile, table, schema, pkey):
		# if dbfile exists, check table exists, check cols
		# else create the table

		self.medoo = Medoo(
			dbtype = 'sqlite',
			database = dbfile,
			check_same_thread = True,
			logging = True
		)
		self.medoo.dialect(DialectCache)
		self.table = table
		self.pkey  = pkey
		self.medoo.query("""
		CREATE TABLE IF NOT EXISTS {} (
			{}
		)
		""".format(table, ',\n'.join(['{} {}'.format(DialectCache.quote(k),v) for k,v in schema.items()])))

	@staticmethod
	def _arrayJoiner(array, delimit = None):
		delimit = delimit or Cache.ARRAY_DELIMIT
		if array:
			return delimit + delimit.join([str(a) for a in array]) + delimit
		else:
			return ''

	@staticmethod
	def _jsonFinder(col, qitem, results):
		ret = []
		for result in results:
			if qitem == result[col]:
				ret.append(result)
		return ret

	@staticmethod
	def _plainFinder(col, qitem, results, case = False):
		ret = []
		if case: qitem = qitem.upper()
		for result in results:
			target = result[col] if not case else result[col].upper()
			if qitem == target:
				ret.append(result)
		return ret

	@staticmethod
	def _arrayFinder(col, qitem, results, case = False):
		ret = []
		if case: qitem = qitem.upper()
		for result in results:
			target = [r.upper() if case else r for r in result[col]]
			if qitem in target:
				ret.append(result)
		return ret

	@staticmethod
	def _uniqueData(data, forceList = False):
		if isinstance(data, list):
			try:
				return list(set(data))
			except TypeError:
				return data
		elif forceList:
			return [data]
		else:
			return data

	@staticmethod
	def _expandLike(data):
		'''
		def expand(d):
			return [
				# // abc
				'%{delimit}{data}'.format(data = d, delimit = Cache.ARRAY_DELIMIT),
				# // abc // def
				'%{delimit}{data}{delimit}%'.format(data = d, delimit = Cache.ARRAY_DELIMIT),
			]
		'''
		ret = []
		for d in Cache._uniqueData(data, True):
			#ret += expand(d)
			ret.append('%{delimit}{data}{delimit}%'.format(data = d, delimit = Cache.ARRAY_DELIMIT))
		return ret

	@staticmethod
	def _queryWhere(cols, data, dummies):
		conditions = []

		if len(cols) == 1:
			conditions.append(Cache._getDummy(cols[0], dummies)['query'](cols[0], data))
		else:
			condkey = 'OR#' + ':'.join(cols)
			conds   = [Cache._getDummy(key, dummies)['query'](key, data) for key in cols]
			conditions.append((condkey, dict(Box(conds))))
		return dict(Box(conditions))

	@staticmethod
	def _find(qitems, results, dummies):
		found = results
		for keys, qitem in qitems.items():
			keys = alwaysList(keys)
			found2 = []
			for key in keys:
				dummy   = Cache._getDummy(key, dummies)
				found2 += dummy['find'](key, qitem, found)
			found = found2
		if found: return found[0]
		#return None

	@staticmethod
	def _checkData(data):
		maxlen = max([len(val) if isinstance(val, list) else 1 for val in data.values()])
		for key, val in data.items():
			if isinstance(val, list) and len(val) > 1 and len(val) != maxlen:
				raise ValueError('Inconsistence of data length.')
			if isinstance(val, list) and len(val) == 1:
				data[key] = val * maxlen
			elif not isinstance(val, list):
				data[key] = [val] * maxlen
		return maxlen

	@staticmethod
	def _getDummy(column, dummies):
		if not dummies or not column in dummies:
			return Cache.DUMMY['plain']
		dummy = dummies[column]
		return dummy if isinstance(dummy, dict) else Cache.DUMMY[dummy]

	@staticmethod
	def _result(data, dummies):
		for key, val in data.items():
			dummy = Cache._getDummy(key, dummies)
			if 'result' in dummy:
				try:
					data[key] = dummy['result'](val)
				except TypeError as ex:
					raise type(ex)(str(ex) + ':\n%s: %s\n' % (key, val))
		return data

	def _queryN(self, columns, data, dummies = None):
		"""
		Query the cached database with N records of data
		@params:
			columns: The columns to select, could be list or comma-concatenated string.
			data:    The data used to query, if query on multiple columns, key could be comma-concatenated string.
			dummies: The dummy functions (query, find) of the columns. If not give, will use self.DUMMY.plain
		@returns:
			results: The records with selected columns
			rest:    The data that records not found
		@examples:
			_queryN('name, symbol', {
				'symbol, alias': ['ABC', 'DEF', 'HIG'],
				'speicies': 'human'
			})
		"""
		columns = alwaysList(columns) + [self.pkey]

		conditions = []
		#datalen = Cache._checkData(data)
		datalen = None
		for keys, dt in data.items():
			datalen  = datalen or len(dt)
			keys     = alwaysList(keys)
			columns += keys
			conditions.append(self._queryWhere(keys, dt, dummies))

		if len(conditions) == 1:
			where = conditions[0]
		else:
			where = {'AND': {}}
			for c in conditions:
				where['AND'].update(c)

		rs = self.medoo.select(self.table, list(set(columns)), where, readonly = False)
		results2 = rs.all() if rs else []
		results3 = [Cache._result(result, dummies) for result in results2 if not any([val is None for val in result.values()])]
		del results2[:]

		results = {}
		for i in range(datalen):
			qitems     = {keys:dt[i] for keys, dt in data.items()}
			results[i] = Cache._find(qitems, results3, dummies)

		rest = {}
		for keys, dt in data.items():
			rest[keys] = [d for i, d in enumerate(dt) if not results[i]]

		return {k:r for k,r in results.items() if r}, rest

	def query(self, columns, data, dummies = None, chunk = 500):
		retall, retrest = {}, {}
		datalen = Cache._checkData(data)
		for i in xrange(0, datalen, chunk):
			datai = {key:val[i:i+chunk] for key, val in data.items()}
			results, rest = self._queryN(columns, datai, dummies)
			retall.update({k+i:v for k,v in results.items()})
			retrest = {key: (val + (retrest[key] if key in retrest else [])) for key, val in rest.items()}
		return retall, retrest

	"""
	data: {
			'id': [4, 5, 6],
			'c2': ['ANAME4', 'ANAME5', 'A6']
	}
	factory: {
		#         update                               insert
		'c2': ((lambda v: Raw('"c2" || \'|%s\'' % v)), lambda v:v)
	}
	"""
	def save(self, data, dummies = None, chunk = 1000):

		datalen = Cache._checkData(data)
		for i in xrange(0, datalen, chunk):
			datai = {key:val[i:i+chunk] for key, val in data.items()}
			self.saveN(datai, dummies)

	def saveN(self, data, dummies):
		# see whether we should update or insert
		pkdata     = data[self.pkey]
		rs         = self.medoo.select(self.table, '*', {self.pkey: pkdata}, readonly = False)
		rsall      = rs.all() if rs else []
		# existing pkeys
		pks        = [r[self.pkey] for r in rsall]
		idxinserts = [i for i, pk in enumerate(pkdata) if pk not in pks]
		idxupdates = [(i, Cache._result(rsall[pks.index(pk)], dummies)) for i, pk in enumerate(pkdata) if pk in pks]

		dataInsert = [
			Box([Cache._getDummy(key, dummies)['insert'](key, val[i]) for key, val in data.items()]) 
			for i in idxinserts
		]
		dataUpdate = [
			(data[self.pkey][i], Box([
				Cache._getDummy(key, dummies)['update'](key, val[i], r[key]) 
				for key, val in data.items() 
				if key != self.pkey
			]))	for i, r in idxupdates
		]

		if dataInsert:
			self.medoo.insert(self.table, *dataInsert, commit = False)
		for pkdataup in dataUpdate:
			pk, dataup = pkdataup
			self.medoo.update(self.table, dataup, {self.pkey: pk}, commit = False)
		self.medoo.commit()



from pyppl import Box
from bioprocs.utils import shell

infile    = {{i.infile | quote}}
outfile   = {{o.outfile | quote}}
inopts    = {{args.inopts | repr}}
n         = {{args.n | repr}}
arsample  = {{args.arsample | quote}}
replace   = {{args.replace | repr}}
keeporder = {{args.keeporder |repr}}
seed      = {{args.seed |repr}}
params    = {{args.params | repr}}
arsample  = shell.Shell(dict(sample = arsample)).sample

if inopts.get('skip', 0):
	shell.head(n = inopts.skip, _ = infile, _stdout = outfile)
	infile_skipped = outfile + '.skipped'
	shell.tail(n = '+' + str(inopts.skip + 1), _ = infile, _stdout = infile_skipped)
	infile = infile_skipped

params._ = infile
params._stdout_ = outfile
params.k = n
params.d = seed
if keeporder:
	params.p = True
if replace:
	params.r = True
else:
	params.o = True

arsample(**params).run()



from pyppl import Box
from bioprocs.utils import logger
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

{% python from os import path %}
{% python from pyppl.utils import alwaysList %}
{% assign cnames = dict(list = list, readlines = readlines,alwaysList = alwaysList, func = lambda x: x) %}
{% assign cntype = lambda x, path = path: 'none' if x is None else 'list' if isinstance(x, (list, tuple)) else 'readlines' if path.isfile(x) else 'func' if x.startswith('lambda') else 'alwaysList' %}
infile  = {{i.infile | quote}}
hfile   = {{i.hfile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts | repr}}
cnames  = {{args.cnames | cnames.get(cntype(args.cnames), repr)}}

if hfile and cnames and not callable(cnames):
	logger.warning('Both header file (hfile) and column names (args.cnames) specified, latter one discarded.')

header = None
if hfile:
	with open(hfile) as f:
		header = [h.strip() for h in f if h.strip()]

reader = TsvReader(infile, **inopts)
if header:
	reader.cnames = header
elif cnames and not callable(cnames):
	reader.cnames = cnames
if callable(cnames):
	reader.cnames = cnames(reader.cnames)

writer = TsvWriter(outfile, delimit = inopts.get('delimit', "\t"))
writer.cnames = reader.cnames
writer.writeHead()
writer.file.write(reader.file.read())

del reader
del writer



import sys
from os import path, makedirs
from pyppl import Box
from bioprocs.utils import mem2, shell2 as shell

infile          = {{i.infile | quote}}
outfile         = {{o.outfile | quote}}
outdir          = {{o.outdir | quote}}
tmpdir          = {{args.tmpdir | quote}}
tool            = {{args.tool | quote}}
snpeff          = {{args.snpeff | quote}}
vep             = {{args.vep | quote}}
gz              = {{args.gz | repr}}
vcfanno         = {{args.vcfanno | quote}}
annovar         = {{args.annovar | quote}}
annovar_convert = {{args.annovar_convert | quote}}
genome          = {{args.genome | quote}}
nthread         = {{args.nthread | repr}}
dbs             = {{args.dbs | repr}}
dbs             = dbs[tool]
snpeffStats     = {{args.snpeffStats | repr}}
params          = {{args.params | repr}}
mem             = {{args.mem | repr}}
tmpdir          = path.join(tmpdir, "{{proc.id}}.{{i.infile | fn}}.{{job.index}}")
if not path.exists(tmpdir):
	makedirs(tmpdir)
if gz:
	outfile = outfile[:-3]

shell.load_config(
	vcfanno         = vcfanno,
	annovar         = annovar,
	annovar_convert = annovar_convert,
	vep             = vep,
	snpeff          = snpeff,
)

def run_snpeff():
	if not path.exists(dbs):
		raise ValueError('Database does not exist: {}'.format(dbs))

	params.dataDir = dbs
	if snpeffStats:
		params.csvStats = path.join(outdir, path.basename(infile) + '.stats.csv')
	else:
		params.noStats = True
	params.i = 'vcf'
	params.o = 'vcf'
	params["Djava.io.tmpdir=%s" % tmpdir] = True
	params._ = [genome, infile]
	params._out = outfile
	params._stderr = sys.stderr
	shell.snpeff.ann(**params)

	if gz:
		shell.bgzip(outfile)

def run_vep():
	if not path.exists(dbs):
		raise ValueError('Database does not exist: {}'.format(dbs))

	params.i        = infile
	params.o        = outfile
	params.format   = 'vcf'
	params.vcf      = True
	params.cache    = True
	params.dir      = dbs
	params.assembly = genome
	shell.fg.vep(**params)

	if gz:
		shell.bgzip(outfile)

def run_annovar():
	if not path.exists(dbs):
		raise ValueError('Database does not exist: {}'.format(dbs))
	import vcf
	avinput              = path.join(outdir, infile + '.avinput')
	avparams             = Box()
	avparams.includeinfo = True
	avparams.allsample   = True
	avparams.withfreq    = True
	avparams.format      = 'vcf4'
	avparams.outfile     = avinput
	avparams._           = infile
	shell.fg.annovar_convert(**avparams)

	params.buildver = genome
	params._        = [avinput, dbs]
	params.outfile  = path.splitext(outfile)[0]
	shell.fg.annovar(**params)

	# convert .variant_function to vcf
	# like snpEff ann, add ANN field to vcf
	def info2dict(info):
		ret   = {}
		infos = info.split(';')
		for i in infos:
			if '=' in i:
				ii = i.split('=')
				if ',' in ii[1]:
					ret[ii[0]] = ii[1].split(',')
				else:
					ret[ii[0]] = ii[1]
			else:
				ret[i] = True
		return ret

	def rs2record(rs):
		r       = rs[0]
		CHROM   = r[2]
		POS     = int(r[3])
		ID      = r[12]
		REF     = r[5]
		ALT     = vcf.model._Substitution(r[6])
		QUAL    = r[15]
		FILTER  = r[16]
		INFO    = info2dict(r[17])
		FORMAT  = r[18]
		samples = r[19:]

		anns = []
		alts = []
		for i, lr in enumerate(rs):
			alts.append(vcf.model._Substitution(lr[6]))
			ann = [lr[6]]
			ann.append(lr[0])  #Annotation
			ann.append('')     #impact
			gene = ''
			dist = '0'
			for genes in lr[1].split(','):
				if genes.startswith('NONE'): continue
				genes = genes.split('(dist=')
				gene  = genes[0]
				if len(genes)>1:
					dist  = genes[1][:-1]
				break
			ann.append(gene)
			ann.append('')    # geneid
			ann.append('')    # Feature type
			ann.append('')    # Feature ID
			ann.append('')    # Transcript biotype
			ann.append('1/1') # Rank / total
			ann.append('')    # HGVS.c
			ann.append('')    # HGVS.p
			ann.append('')    # cDNA_position
			ann.append('')    # CDS_position
			ann.append('')    # Protein_position
			ann.append(dist)  # Distance to feature
			ann.append('')  # Errors, Warnings or Information messages
			anns.append ('|'.join(ann))
		INFO['ANN'] = anns

		record  = vcf.model._Record(CHROM, POS, ID, REF, alts, QUAL, FILTER, INFO, FORMAT, snames)
		record.samples = reader._parse_samples (samples, FORMAT, record)
		return record

	reader = vcf.Reader(filename="{{i.infile}}")
	reader.infos["ANN"] = vcf.parser._Info("ANN", 1, "String", "Annotation by ANNOVAR", "", "")
	snames = {v:k for k,v in enumerate(reader.samples)}
	writer = vcf.Writer(open(outfile, 'w'), reader)
	f2conv = "{{o.outfile | prefix}}.variant_function"
	lastvid= ''
	lastr  = []
	with open (f2conv) as f:
		for line in f:
			line = line.strip("\r\n")
			if not line: continue

			parts = line.split("\t")
			varid = parts[2] + '|' + parts[3] + '|' + parts[12] + '|' + parts[5]

			if lastvid != varid and lastvid:
				record  = rs2record(lastr)
				writer.write_record(record)
				lastvid = varid
				lastr   = [parts]
			else:
				lastvid = varid
				lastr.append (parts)

	record  = rs2record(lastr)
	writer.write_record(record)
	writer.close()

	if gz:
		shell.bgzip(outfile)

def run_vcfanno():
	# compose toml file
	toml = path.join(outdir, 'config.toml')
	with open(toml, 'w') as f:
		for config in dbs:
			f.write("\n[[annotation]]\n")
			for key, val in config.items():
				f.write('{key}={val!r}\n'.format(key = key, val = val))
	params.p       = nthread
	params._       = [toml, infile]
	params._out    = outfile
	params._stderr = sys.stderr
	shell.vcfanno(**params)

	if gz:
		shell.bgzip(outfile)

tools = dict(
	snpeff  = run_snpeff,
	vep     = run_vep,
	annovar = run_annovar,
	vcfanno = run_vcfanno
)


try:
	tools[tool]()
except KeyError:
	raise ValueError('Tool {!r} not supported yet.'.format(tool))
except:
	raise
finally:
	shell.rm_rf(tmpdir)



from pyppl import Aggr
from bioprocs.common import pFile2Proc
from bioprocs.rnaseq import pExprDir2Matrix, pRNASeqDEG, pExprPlot
from bioprocs.marray import pCELDir2Matrix, pMArrayDEG
from bioprocs.resource import pTxt
from bioprocs.gsea import pExprMat2GCT, pSampleinfo2CLS, pGSEA, pEnrichr

"""
@name:
	aExpDir2Deg
@description:
	From expfiles to degs with sample info file.
@depends:
	```
	pExprDir2Matrix[*] \
	                      pRNASeqDEG[!]
	pSampleInfo[*]     /
	```
"""
aExpDir2Deg = Aggr(
	pExprDir2Matrix,
	pFile2Proc.copy(id = 'pSampleInfo'),
	pRNASeqDEG,
	depends = False
)
# Dependences
aExpDir2Deg.starts             = aExpDir2Deg.pExprDir2Matrix, aExpDir2Deg.pSampleInfo
aExpDir2Deg.ends               = aExpDir2Deg.pRNASeqDEG
aExpDir2Deg.pRNASeqDEG.depends = aExpDir2Deg.pExprDir2Matrix, aExpDir2Deg.pSampleInfo
# Input
aExpDir2Deg.pRNASeqDEG.input = lambda ch1, ch2: \
	ch1.colAt(0).repRow(ch2.length()).cbind(ch2.repRow(ch1.length()))
# Args
aExpDir2Deg.pExprDir2Matrix.args.boxplot  = True
aExpDir2Deg.pExprDir2Matrix.args.heatmap  = True
aExpDir2Deg.pExprDir2Matrix.args.histplot = True
aExpDir2Deg.pRNASeqDEG.args.maplot       = True
aExpDir2Deg.pRNASeqDEG.args.heatmap      = True


"""
@name:
	aExpDir2DegGSEA
@description:
	From expfiles to degs with sample info file and do GSEA.
@depends:
	```
	                  /   pRNASeqDEG[!]  -> pEnrichr[!]
	pExprDir2Matrix[*] \
	                      pExprMat2GCT     \
	pSampleInfo[*]     /                   \
	                   \  pSampleinfo2CLS --  pGSEA[!]
	pGMTFetcher[*]     ____________________/
	```
"""
aExpDir2DegGSEA = Aggr(
	pExprDir2Matrix,
	pFile2Proc.copy(id = 'pSampleInfo'),
	pTxt.copy(id = 'pGMTFetcher'),
	pExprPlot,
	pExprMat2GCT,
	pSampleinfo2CLS,
	pGSEA,
	pRNASeqDEG,
	pEnrichr,
	depends = False
)
# Default input:
aExpDir2DegGSEA.pGMTFetcher.input = ['KEGG_2016_gmt']
# Dependences
aExpDir2DegGSEA.starts                  = aExpDir2DegGSEA.pExprDir2Matrix,  \
										  aExpDir2DegGSEA.pSampleInfo,     \
										  aExpDir2DegGSEA.pGMTFetcher
aExpDir2DegGSEA.ends                    = aExpDir2DegGSEA.pRNASeqDEG,      \
										  aExpDir2DegGSEA.pGSEA,           \
										  aExpDir2DegGSEA.pEnrichr
aExpDir2DegGSEA.pExprMat2GCT.depends     = aExpDir2DegGSEA.pExprDir2Matrix
aExpDir2DegGSEA.pSampleinfo2CLS.depends = aExpDir2DegGSEA.pSampleInfo
aExpDir2DegGSEA.pGSEA.depends           = aExpDir2DegGSEA.pExprMat2GCT,     \
										  aExpDir2DegGSEA.pSampleinfo2CLS, \
										  aExpDir2DegGSEA.pGMTFetcher
aExpDir2DegGSEA.pRNASeqDEG.depends      = aExpDir2DegGSEA.pExprDir2Matrix,  \
										  aExpDir2DegGSEA.pSampleInfo
aExpDir2DegGSEA.pEnrichr.depends        = aExpDir2DegGSEA.pRNASeqDEG
# Input
aExpDir2DegGSEA.pGSEA.input      = lambda ch1, ch2, ch3: \
	ch1.repRow(ch2.length() * ch3.length())              \
	   .cbind(ch2.repRow(ch1.length() * ch3.length()))   \
	   .cbind(ch3.repRow(ch1.length() * ch2.length())) 
aExpDir2DegGSEA.pRNASeqDEG.input = lambda ch1, ch2: \
	ch1.colAt(0).repRow(ch2.length()).cbind(ch2.repRow(ch1.length()))
# Args
aExpDir2DegGSEA.pExprDir2Matrix.args.boxplot  = True
aExpDir2DegGSEA.pExprDir2Matrix.args.heatmap  = True
aExpDir2DegGSEA.pExprDir2Matrix.args.histplot = True
aExpDir2DegGSEA.pRNASeqDEG.args.maplot       = True
aExpDir2DegGSEA.pRNASeqDEG.args.heatmap      = True
aExpDir2DegGSEA.pGMTFetcher.args.header      = False

"""
@name:
	aRnaseqExpMat2DegGSEA
@description:
	From expfiles to degs with sample info file and do GSEA.
@depends:
	```
	                  /   pRNASeqDEG[!]  -> pEnrichr[!]
	pExpMat[*]        \
	                      pExprMat2GCT     \
	pSampleInfo[*]     /                   \
	                   \  pSampleinfo2CLS --  pGSEA[!]
	pGMTFetcher[*]     ____________________/
	```
"""
aRnaseqExpMat2DegGSEA = Aggr(
	pFile2Proc.copy(id = 'pExpMat'),
	pFile2Proc.copy(id = 'pSampleInfo'),
	pTxt.copy(id = 'pGMTFetcher'),
	pExprMat2GCT,
	pSampleinfo2CLS,
	pGSEA,
	pRNASeqDEG,
	pEnrichr,
	depends = False
)
# Default input:
aRnaseqExpMat2DegGSEA.pGMTFetcher.input = ['KEGG_2016_gmt']
# Dependences
aRnaseqExpMat2DegGSEA.starts                  = aRnaseqExpMat2DegGSEA.pExpMat,  \
												aRnaseqExpMat2DegGSEA.pSampleInfo,     \
												aRnaseqExpMat2DegGSEA.pGMTFetcher
aRnaseqExpMat2DegGSEA.ends                    = aRnaseqExpMat2DegGSEA.pRNASeqDEG,      \
												aRnaseqExpMat2DegGSEA.pGSEA,           \
												aRnaseqExpMat2DegGSEA.pEnrichr
aRnaseqExpMat2DegGSEA.pExprMat2GCT.depends     = aRnaseqExpMat2DegGSEA.pExpMat
aRnaseqExpMat2DegGSEA.pSampleinfo2CLS.depends = aRnaseqExpMat2DegGSEA.pSampleInfo
aRnaseqExpMat2DegGSEA.pGSEA.depends           = aRnaseqExpMat2DegGSEA.pExprMat2GCT,     \
												aRnaseqExpMat2DegGSEA.pSampleinfo2CLS, \
												aRnaseqExpMat2DegGSEA.pGMTFetcher
aRnaseqExpMat2DegGSEA.pRNASeqDEG.depends      = aRnaseqExpMat2DegGSEA.pExpMat,  \
												aRnaseqExpMat2DegGSEA.pSampleInfo
aRnaseqExpMat2DegGSEA.pEnrichr.depends        = aRnaseqExpMat2DegGSEA.pRNASeqDEG
# Input
aRnaseqExpMat2DegGSEA.pGSEA.input      = lambda ch1, ch2, ch3: \
	ch1.repRow(ch2.length() * ch3.length())              \
	   .cbind(ch2.repRow(ch1.length() * ch3.length()))   \
	   .cbind(ch3.repRow(ch1.length() * ch2.length())) 
aRnaseqExpMat2DegGSEA.pRNASeqDEG.input = lambda ch1, ch2: \
	ch1.colAt(0).repRow(ch2.length()).cbind(ch2.repRow(ch1.length()))
# Args
aRnaseqExpMat2DegGSEA.pRNASeqDEG.args.maplot       = True
aRnaseqExpMat2DegGSEA.pRNASeqDEG.args.heatmap      = True
aRnaseqExpMat2DegGSEA.pGMTFetcher.args.header      = False



"""
@name:
	aCELDir2DEG
@description:
	From celfils to degs with sample info file and do GSEA.
@procs:
	`pCELDir2Matrix[*]`: Convert CEL files to matrix
	`pSampleInfo[*]`   : Receive sample information file. Copied from `pFile2Proc`
	`pGMTFetcher[*]`   : Download the GMT file for GSEA. Copied from `pTxt`
	`pExprPlot[!]`     : Plot the expression distributions.
	`pExprMat2GCT`     : Convert expression matrix to GCT file for GSEA
	`pGSEA[!]`         : Do GSEA
	`pMArrayDEG[!]`    : Call DEGs
	`pEnrichr[!]`      : Do enrichment with DEGs
@modules
	`exprplot`: Plot expression distributions from expression matrix
		- `pCELDir2Matrix[*]` -> `pExprPlot[!]`
	`gsea`: Do GSEA
		- `pCELDir2Matrix[*]` -> `pExprMat2GCT`
		- `pSampleInfo[*]`    -> `pSampleinfo2CLS`
		- `pExprMat2GCT`, `pCELDir2Matrix[*]`, `pGMTFetcher` -> `pGSEA[!]`
	`enrichr`: Do enrichr
		- `pCELDir2Matrix[*]`, `pSampleInfo[*]` -> `pMArrayDEG[!]` -> `pEnrichr[!]`
"""
aCELDir2DEG = Aggr(
	pCELDir2Matrix,
	pFile2Proc.copy(id = 'pSampleInfo'),
	pTxt.copy(id = 'pGMTFetcher'),
	pExprPlot,
	pExprMat2GCT,
	pSampleinfo2CLS,
	pGSEA,
	pMArrayDEG,
	pEnrichr,
	depends = False
)
# Defaults
aCELDir2DEG.starts = 'pCELDir2Matrix, pSampleInfo'
aCELDir2DEG.ends   = 'pMArrayDEG'

aCELDir2DEG['pGMTFetcher'].input        = ['KEGG_2016_gmt']
aCELDir2DEG['pMArrayDEG'].args.maplot   = True
aCELDir2DEG['pMArrayDEG'].args.heatmap  = True
aCELDir2DEG['pGMTFetcher'].args .header = False
# Depends
aCELDir2DEG['pMArrayDEG', ].depends = ['pCELDir2Matrix, pSampleInfo']
# Input adjustments
aCELDir2DEG['pGSEA'].input = lambda ch1, ch2, ch3:       \
	ch1.repRow(ch2.length() * ch3.length())              \
	   .cbind(ch2.repRow(ch1.length() * ch3.length()))   \
	   .cbind(ch3.repRow(ch1.length() * ch2.length())) 
aCELDir2DEG['pMArrayDEG'].input = lambda ch1, ch2:       \
	ch1.colAt(0).repRow(ch2.length()).cbind(ch2.repRow(ch1.length()))
# Configs
aCELDir2DEG.module('exprplot', depends = {
	'pExprPlot': 'pCELDir2Matrix'
}, ends = 'pExprPlot')
aCELDir2DEG.module('gsea', 'pGMTFetcher', {
	'pGSEA'          : 'pExprMat2GCT, pSampleinfo2CLS, pGMTFetcher',
	'pExprMat2GCT'   : 'pCELDir2Matrix',
	'pSampleinfo2CLS': 'pSampleInfo'
}, 'pGSEA')
aCELDir2DEG.module('enrichr', depends = {
	'pEnrichr': 'pMArrayDEG'
}, ends = 'pEnrichr')
# Enable all functions
aCELDir2DEG.on()




import types, tempfile

class TsvRecord(object):
	__slots__ = ('__keys', '__vals')

	def __init__(self, vals = None, keys = None):
		"""
		r = TsvRecord([1,2,3,4,5])
		r.__vals == [1,2,3,4,5]
		r.__keys == None
		"""
		self.__vals = vals or []
		if keys:
			self.__keys = dict(zip(keys, range(len(keys))))
			if len(self.__keys) != len(self.__vals):
				raise ValueError("Unequal length of keys and values. Make sure you don't have duplicated keys.")
		else:
			self.__keys = None

	def attachKeys(self, keys):
		assert len(keys) == len(self.__vals)
		self.__keys = dict(zip(keys, range(len(keys))))

	def keys(self):
		if not self.__keys:
			return range(len(self.__vals))
		return sorted(self.__keys, key=self.__keys.get)

	def values(self):
		return self.__vals

	def items(self):
		if not self.__keys:
			return enumerate(self.__vals)
		return zip(list(self.keys()), list(self.__vals))

	def __repr__(self):
		return '<TsvRecord: {!r}>'.format(dict(self.items()))

	def __getitem__(self, key):
		if isinstance(key, (slice, int)):
			return self.__vals[key]

		if self.__keys and key in self.__keys:
			return self.__vals[self.__keys[key]]

		raise KeyError("Record contains no '{}' field.".format(key))

	def __getattr__(self, key):
		if str(key).startswith('__') or str(key).startswith('_TsvRecord'):
			return super(TsvRecord, self).__getattr__(key)
		return self[key]

	def __setitem__(self, key, value):
		if isinstance(key, int):
			if key > len(self) or key < 0:
				raise IndexError('Index out of range: {}'.format(key))
			self.__vals[key] = value
		elif self.__keys and key in self.__keys:
			self.__vals[self.__keys[key]] = value
		else:
			self.__keys = self.__keys or {}
			self.__keys[key] = len(self)
			self.__vals.append(value)

	def __len__(self):
		return len(self.__vals)

	def __setattr__(self, key, value):
		if str(key).startswith('__') or str(key).startswith('_TsvRecord'):
			super(TsvRecord, self).__setattr__(key, value)
		else:
		  self[key] = value

	def get(self, key, default=None):
		"""Returns the value for a given key, or default."""
		try:
			return self[key]
		except KeyError:
			return default

	def __eq__(self, other):
		return self.keys() == other.keys() and self.values() == other.values()

	def __ne__(self, other):
		return not self.__eq__(other)

	def asDict(self, ordered=False):
		"""Returns the row as a dictionary, as ordered."""
		return OrderedDict(self.items()) if ordered else dict(items.items())

	def __contains__(self, key):
		return self.__keys and key in self.__keys

	def __delitem__(self, key):
		if not self.__keys:
			del self.__vals[key]
		elif key in self.__keys:
			del self.__vals[self.__keys[key]]
			del self.__keys[key]
		else:
			del self.__keys[list(self.keys())[key]]
			del self.__vals[key]

class TsvReader(object):

	def __init__(self,
		infile,
		delimit = '\t',
		comment = '#',
		skip    = 0,
		cnames  = True, # "False": no head; "None"/"True": split first line with delimit, "Callback": get head for first line in your way
		attach  = True,
		row     = None, # row factory
		cname0  = "ROWNAME"):
		openfunc = open
		# in case infile is a Pathlib.Path object
		infile = str(infile)
		if infile.endswith('.gz'):
			import gzip
			openfunc = gzip.open

		self.file    = openfunc(infile, errors='replace')
		self.delimit = delimit
		self.comment = comment
		self.attach  = attach
		self.row     = row
		self.tell    = 0

		if skip > 0:
			for _ in range(skip):
				self.file.readline()

		while True:
			tell = self.file.tell()
			line = self.file.readline()
			if comment and line.startswith(comment):
				continue
			self.file.seek(tell)
			break

		headline = self.file.readline() if cnames is not False else ''
		if callable(cnames):
			self.cnames = cnames(headline)
		elif headline:
			if comment and headline.startswith(comment):
				headline = headline[1:].lstrip()
			self.cnames = headline.rstrip('\n').split(delimit)
		else:
			self.cnames = []
		# try to add "cname0" as column name
		tell  = self.file.tell()
		firstline = self.file.readline().rstrip('\n')
		ncols = len(firstline.split(delimit))

		if firstline and self.cnames and len(self.cnames) == ncols - 1:
			self.cnames.insert(0, cname0)
		if firstline and self.cnames and len(self.cnames) != ncols:
			raise ValueError('Not a valid tsv file. Head has %s columns, while first line has %s.' % (len(self.cnames), ncols))

		self.file.seek(tell)
		self.tell = tell
		self.meta = self.cnames

	def next(self):
		line = self.file.readline()
		line = line.rstrip('\n')
		# empty lines not allowed
		if not line: raise StopIteration()
		record = TsvRecord(line.split(self.delimit))
		if self.attach and self.cnames:
			record.attachKeys(self.cnames)
		if callable(self.row):
			return self.row(record)
		return record

	def dump(self, col = None):
		if col is None:
			return list(self)
		if not isinstance(col, list):
			return [r[col] for r in self]
		return [tuple(r[c] for c in col) for r in self]

	def __next__(self):
		return self.next()

	def rewind(self):
		self.file.seek(self.tell)

	def __iter__(self):
		return self

	def __del__(self):
		self.close()

	def close(self):
		if hasattr(self, 'file') and self.file:
			self.file.close()

class TsvWriter(object):
	def __init__(self, outfile = None, delimit = '\t', append = False):
		openfunc = open
		outfile = str(outfile) # support pathlib
		if outfile and outfile.endswith('.gz'):
			import gzip
			openfunc = gzip.open

		self.delimit = delimit
		self.cnames  = []
		if outfile:
			self.file     = openfunc(outfile, 'w' if not append else 'a')
			self.filename = outfile
		else:
			self.file     = tempfile.NamedTemporaryFile(delete = False)
			self.filename = self.file.name
		self.meta    = self.cnames # alias

	def writeHead(self, callback = True):
		if not self.cnames:
			return
		if callback and callable(callback):
			head = callback(self.cnames)
			if isinstance(head, list):
				head = self.delimit.join(head)
			self.file.write(head + "\n")
		elif callback:
			head = self.delimit.join(self.cnames)
			self.file.write(head + "\n")

	def write(self, record):
		if isinstance(record, list) or isinstance(record, types.GeneratorType):
			self.file.write(self.delimit.join(str(v) for v in record) + '\n')
		elif isinstance(record, TsvRecord):
			if not self.cnames:
				self.write(record.values())
			else:
				self.write(record[n] for n in self.cnames)
		else:
			self.file.write(str(record))

	def __del__(self):
		self.close()

	def close(self):
		if hasattr(self, 'file') and self.file:
			self.file.close()

class TsvJoin(object):

	@staticmethod
	def compare(a, b, reverse = False):
		if not reverse:
			return 0 if a < b else 1 if a > b else -1
		else:
			return 0 if a > b else 1 if a < b else -1

	@staticmethod
	def _debug(rows, m):
		from sys import stderr
		for i, row in enumerate(rows):
			stderr.write("- File {}: {}\n".format(i+1, row))
		if m < 0:
			stderr.write("- Matched\n")
		else:
			stderr.write("- File {} is behand\n".format(i+1))
		stderr.write('-' * 80 + "\n")

	def __init__(self, *files, **inopts):
		inopts_default = dict(
			delimit = '\t',
			comment = '#',
			skip    = 0,
			cnames  = True,
			attach  = False,
			row     = None,
			cname0  = "ROWNAME"
		)
		inopts_multi = {}
		self.length = len(files)
		for key, val in inopts.items():
			if not isinstance(val, list):
				inopts_multi[key] = [val] * self.length
			elif len(val) < self.length:
				inopts_multi[key] = val + [inopts_default[key]] * (self.length - len(val))
			else:
				inopts_multi[key] = val
		inopts = []
		for i in range(self.length):
			inopts.append({k:v[i] for k,v in inopts_multi.items()})
		self.readers = [TsvReader(f, **inopts[i]) for i,f in enumerate(files)]
		self.debug   = False

	def _defaultMatch(self, *rows):
		data = [row[0] for row in rows]
		mind = min(data)
		return -1 if data.count(mind) == self.length else data.index(mind)

	def join(self, do, outfile, match = None, outopts = None):
		outopts = outopts or {}
		outopts_default = dict(
			delimit = "\t",
			append  = False
		)
		outopts_default.update(outopts)
		outopts = outopts_default

		cnames = False
		if 'cnames' in outopts:
			cnames = outopts['cnames']
			del outopts['cnames']

		headCallback = None
		if 'headCallback' in outopts:
			headCallback = outopts['headCallback']
			del outopts['headCallback']

		out = TsvWriter(outfile, **outopts)
		out.cnames = cnames if isinstance(cnames, list) \
			else sum((reader.cnames for reader in self.readers if reader.cnames), []) if cnames \
			else []
		if out.cnames:
			out.writeHead(headCallback)

		match = match or self._defaultMatch
		rows = [None] * self.length
		while True:
			try:
				for i, row in enumerate(rows):
					rows[i] = row or next(self.readers[i])
				m = match(*rows)
				if self.debug:
					self._debug(rows, m)
				if m < 0: # matched
					do(out, *rows)
					m = 0
				rows[m] = None
			except StopIteration:
				break
			except Exception:
				from sys import stderr
				from traceback import format_exc
				info = format_exc().splitlines()
				info.append("With rows:")
				info.extend(["- {}".format(r) for r in rows])
				stderr.write("\n".join(info) + "\n\n")
				rows = [None] * self.length
				continue
		out.close()





import json, requests, math, re, random
from sys import stderr
from os import path
from hashlib import md5
from mygene import MyGeneInfo
from collections import OrderedDict
{% if args.enrplot %}
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from matplotlib import gridspec
from matplotlib import patches
{% endif %}
from pyppl import Box
from bioprocs.utils import alwaysList, logger
from bioprocs.utils.gene import genenorm
from bioprocs.utils.tsvio import TsvReader, TsvWriter, TsvRecord

infile   = {{i.infile | quote}}
inopts   = {{args.inopts}}
genecol  = {{args.genecol | quote}}
cachedir = {{args.cachedir | quote}}

{% if args.norm %}
symfile = '{{o.outdir}}/{{i.infile | bn}}.symbol'

genenorm(
	infile   = infile,
	outfile  = symfile,
	inopts   = inopts,
	outopts  = {'head': True, 'query': False},
	genecol  = genecol,
	cachedir = cachedir
)

infile = symfile
{% endif %}

genestats  = {}
regstats   = {}
relations  = {}

reader = TsvReader(infile, **inopts)

for r in reader:
	if not r.Regulator in relations:
		relations[r.Regulator] = {}
	genestats[r.Target]   = ''
	regstats[r.Regulator] = ''
	relations[r.Regulator][r.Target] = '+'
	if 'TargetStatus' in r:
		genestats[r.Target] = r.TargetStatus
	if 'RegulatorStatus' in r:
		regstats[r.Regulator] = r.RegulatorStatus
	if 'Relation' in r:
		relations[r.Regulator][r.Target] = r.Relation
if not genestats:
	raise ValueError('No genes found.')

## upload
ENRICHR_URL = 'http://amp.pharm.mssm.edu/Enrichr/addList'
genes_str   = "\n".join(genestats.keys())
description = '{{i.infile | fn}}'
payload = {
    'list': (None, genes_str),
    'description': (None, description)
}

response = requests.post(ENRICHR_URL, files=payload)
if not response.ok:
    raise Exception('Error analyzing gene list')

data = json.loads(response.text)

## do enrichment
dbs = alwaysList({{args.dbs | quote}})

ENRICHR_URL = 'http://amp.pharm.mssm.edu/Enrichr/enrich'
query_string = '?userListId=%s&backgroundType=%s'

head = ["Rank", "Term", "Pval", "Zscore", "CombinedScore", "OverlappingGenes", "AdjustedPval", "OldPval", "OldAdjustedPval"]

enrn = {{args.enrn}}
for db in dbs:
	user_list_id = data['userListId']
	gene_set_library = db
	response = requests.get(
		ENRICHR_URL + query_string % (user_list_id, gene_set_library)
	)
	if not response.ok:
		raise Exception('Error fetching enrichment results against %s' % db)

	data    = json.loads(response.text)
	data    = data[db]
	d2plot  = []
	outfile = "{{o.outdir}}/{{i.infile | fn}}-%s.txt" % db
	writer  = TsvWriter(outfile)
	writer.meta.add(*head)
	writer.writeHead()
	# for network plot
	path2genes = {}
	path2pvals = {}
	genestats2 = {}
	for ret in data[:enrn]:
		r = TsvRecord()
		r.Rank             = ret[0]
		r.Term             = ret[1]
		r.Pval             = ret[2]
		r.Zscore           = ret[3]
		r.CombinedScore    = ret[4]
		r.OverlappingGenes = '|'.join(ret[5])
		r.AdjustedPval     = ret[6]
		r.OldPval          = ret[7]
		r.OldAdjustedPval  = ret[8]
		writer.write(r)

		if {{args.rmtags}} and "_" in ret[1]: ret[1] = ret[1].split('_')[0]
		if len(path2genes) < {{args.netn}}:
			path2genes[r.Term] = r.OverlappingGenes.split('|')
			for g in path2genes[r.Term]:
				if g in genestats:
					genestats2[g] = genestats[g]
				else:
					genestats2[g] = ''
					logger.warn('{gene} not found in query set, you probably need to normalize the gene names.'.format(gene = g))
			path2pvals[r.Term] = r.Pval
		d2plot.append (ret)
	writer.close()

	if {{args.netplot}}:
		# cleanup data
		# map genes back
		#genestats  = {}
		#regstats   = {}
		#relations  = {}
		regstats2   = {}
		relations2  = {}
		for reg, rel in relations.items():
			foundreg = False
			for g, stat in rel.items():
				#if not g in genemap: continue
				#g = genemap[g]
				if not g in genestats2: continue
				foundreg = True
				if not reg in relations2:
					relations2[reg] = {}
				relations2[reg][g] = stat
			if foundreg:
				regstats2[reg] = regstats[reg]

		genemarks  = {g:0 for g in genestats2.keys()}
		plainlinks = []
		for pathw, genes in path2genes.items():
			newgenes = []
			for g in genes:
				if genemarks[g]:
					newg = g + '__' + str(genemarks[g])
					newgenes.append(g + '__' + str(genemarks[g]))
					plainlinks.append((g, newg))
					genestats2[newg] = ''
				else:
					newgenes.append(g)
				genemarks[g] += 1
			path2genes[pathw] = newgenes

		# prepare dot:
		dotstr = ['digraph G {']
		# regulators
		regupcolors = ['#CCFF99', '#99FF33', '#66CC00', '#336600', '#99CC66', '#669933', '#99FFCC', '#33FF99', '#00CC66', '#006633', '#66CC99', '#339966'] # '+'
		regdncolors = ['#FF99CC', '#FF3399', '#CC0066', '#660033', '#CC6699', '#993366', '#FFCC99', '#FF9933', '#CC6600', '#663300', '#CC9966', '#996633'] # '-'
		regplcolors = ['#3399FF'] # ''
		regcolors   = {} # for edge
		for reg, stat in regstats2.items():
			color = random.choice(regupcolors if stat == '+' else regdncolors if stat == '-' else regplcolors)
			regcolors[reg] = color
			dotstr.append('  "%s" [shape=box fontsize=60 height=2 style=filled fillcolor="%s"]' % (reg, color))
		# genes
		geneupcolor = '#00CC00'
		genedncolor = '#FF3333'
		geneplcolor = '#AAAAAA'
		for gene, stat in genestats2.items():
			# fake gene
			if '__' in gene:
				dotstr.append('  "%s" [shape=circle fontsize=50 width=2 height=2 fixedsize=true label="%s" style=filled color="#333333" fillcolor="#DDDDDD"]' % (gene, gene.split('__')[0]))
			else:
				dotstr.append('  "%s" [shape=circle fontsize=50 width=2 height=2 fixedsize=true style=filled color="#333333" fillcolor="%s"]' % (gene, geneupcolor if stat == '+' else genedncolor if stat == '-' else geneplcolor))
		# regulation links
		for reg, rel in relations2.items():
			for gene, stat in rel.items():
				dotstr.append('  "%s" -> "%s" [arrowsize=2 penwidth=8 arrowhead=%s weight=2 color="%s"]' % (reg, gene, ('tee' if stat == '-' else 'normal'), regcolors[reg]))
		# plain links:
		for g1, g2 in plainlinks:
			dotstr.append('  "%s" -> "%s" [dir=none style=dashed width=1 weight=1]' % (g1, g2))
		# subgraphs
		for pathw, genes in path2genes.items():
			dotstr.append('  subgraph cluster_%s { ' % re.sub(r'[^\w]', '', pathw))
			if path2pvals[pathw] < 0.05:
				dotstr.append('    style = filled;')
				dotstr.append('    color = "#FFCCCC";')
			dotstr.append('    fontsize = 60;')
			dotstr.append('    label = "%s (p=%.2E)";' % (pathw, path2pvals[pathw]))
			for g in genes:
				dotstr.append('    "%s";' % g)
			dotstr.append('  }')
		dotstr.append('}')

		from graphviz import Source
		dotfile = '{{o.outdir}}/%s.net.dot' % db
		src = Source('\n'.join(dotstr), filename = dotfile, format = 'svg', engine = 'fdp')
		src.render()

	if {{args.enrplot}}:
		#d2plot   = sorted (d2plot, cmp=lambda x,y: 0 if x[2] == y[2] else (-1 if x[2] < y[2] else 1))
		plotfile = "{{o.outdir}}/{{i.infile | fn}}-%s.png" % db
		gs = gridspec.GridSpec(1, 2, width_ratios=[3, 7])
		rownames = [r[1] if len(r[1])<=40 else r[1][:40] + ' ...' for r in d2plot]
		rnidx    = range (len (rownames))
		ax1 = plt.subplot(gs[0])
		plt.title ("{{args.title}}".replace("{db}", db), fontweight='bold')

		ax1.xaxis.grid(alpha=.6, ls = '--', zorder = -99)
		plt.subplots_adjust(wspace=.01, left=0.5)
		ax1.barh(rnidx, [len(r[5]) for r in d2plot], color='blue', alpha=.6)
		plt.yticks (rnidx, rownames)
		ax1.yaxis.set_ticks_position('none')
		ax1.tick_params(axis='x', colors='blue')
		ax1.spines['top'].set_visible(False)
		ax1.spines['left'].set_visible(False)
		ax1.spines['right'].set_visible(False)
		ax1.spines['bottom'].set_linewidth(1)
		ax1.invert_xaxis()
		ax1.invert_yaxis()
		xticks = ax1.xaxis.get_major_ticks()
		xticks[0].label1.set_visible(False)

		ax2 = plt.subplot(gs[1])
		ax2.xaxis.grid(alpha=.6, ls = '--', zorder = -99)
		ax2.barh(rnidx, [-math.log(r[2], 10) for r in d2plot], color='red', alpha = .6)
		for i, r in enumerate(d2plot):
			t  = str("%.2E" % r[2])
			tx = 0.1
			ty = i + 0.1
			ax2.text(tx, ty, t, fontsize=8)
		ax2.tick_params(axis='x', colors='red')
		ax2.spines['top'].set_visible(False)
		ax2.spines['left'].set_visible(False)
		ax2.spines['right'].set_visible(False)
		ax2.spines['bottom'].set_linewidth(1)
		ax2.invert_yaxis()
		plt.yticks([])
		ng_patch = patches.Patch(color='blue', alpha=.6, label='# overlapped genes')
		pv_patch = patches.Patch(color='red', alpha = .6, label='-log(p-value)')
		plt.figlegend(handles=[ng_patch, pv_patch], labels=['# overlapped genes', '-log(p-value)'], loc="lower center", ncol=2, edgecolor="none")
		plt.savefig(plotfile, dpi=300)



from os import path
from sys import stderr
from pyppl import Box
from bioprocs.utils import parallel, shell2 as shell
from bioprocs.utils.reference import vcfIndex

infile   = {{i.infile | quote}}
prefix   = {{i.infile | fn2 | quote}}
outdir   = {{o.outdir | quote}}
samples  = {{i.samples | quote}}
tool     = {{args.tool | quote}}
bcftools = {{args.bcftools | quote}}
gatk     = {{args.gatk | quote}}
tabix    = {{args.tabix | quote}}
ref      = {{args.ref | quote}}
params   = {{args.params | repr}}
nthread  = {{args.nthread | repr}}

shell.load_config(bcftools = bcftools, gatk = gatk)
vcfIndex(infile, tabix = tabix)

allsamples = shell.bcftools.query(l = infile).splitlines()
allsamples = [s.strip() for s in allsamples if s.strip()]
if samples:
	with open(samples) as f:
		samples = f.readlines()
	samples = list(set(allsamples) & set(samples))
else:
	samples = allsamples

def run_bcftools_one(sample):
	shell.fg.bcftools.view(_ = infile, s = sample, o = path.join(outdir, '{}-{}.vcf'.format(prefix, sample)), **params)

def run_bcftools():
	parallel.Parallel(nthread).run(run_bcftools_one, [(sample,) for sample in samples])

def run_awk_one(sample, index, awkfile):
	shell.awk(
		v = ["sample={!r}".format(sample), "index={}".format(index + 10)],
		_stderr = stderr,
		f = awkfile,
		_ = infile,
		_out = path.join(outdir, '{}-{}.vcf'.format(prefix, sample)))

def run_awk():
	# write the awk script
	awkfile = path.join(outdir, 'vcfsample.awk')
	awkfh   = open(awkfile, 'w')
	awkfh.write("""
BEGIN {
	OFS="\\t"
}
$0 ~ "^##" {
	print
}
$0 ~ "^#CHROM" {
	print "#CHROM\\tPOS\\tID\\tREF\\tALT\\tQUAL\\tFILTER\\tINFO\\tFORMAT\\t"sample
}
$0 !~ "^#" {
	print $1,$2,$3,$4,$5,$6,$7,$8,$9,$index
}
""")
	awkfh.close()
	parallel.Parallel(nthread).run(run_awk_one, [(sample, i, awkfile) for i, sample in enumerate(samples)])

def run_gatk_one(sample):
	shell.fg.gatk(
		R = ref,
		V = infile,
		o = path.join(outdir, '{}-{}.vcf'.format(prefix, sample)),
		sample_name = sample,
		T = 'SelectVariants',
		excludeNonVariants = True,
		**params
	)

def run_gatk():
	parallel.Parallel(nthread).run(run_gatk_one, [(sample, ) for sample in samples])

tools = dict(bcftools = run_bcftools, awk = run_awk, gatk = run_gatk)

try:
	tools[tool]()
except KeyError:
	raise ValueError('Tool {!r} not supported yet.'.format(tool))



from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.parallel import Parallel
from bioprocs.utils.reference import vcfIndex

invcfs    = {{i.infiles | repr}}
outfile   = {{o.outfile | quote}}
nthread   = {{args.nthread | int}}
joboutdir = {{job.outdir | quote}}
vcftools  = {{args.vcftools | quote}}
bcftools  = {{args.bcftools | quote}}
gatk      = {{args.gatk | quote}}
tabix     = {{args.tabix | quote}}
ref       = {{args.ref | quote}}
params    = {{args.params}}
params    = params if isinstance(params, Box) else Box(params)
tool      = {{args.tool | quote}}
gz        = {{args.gz | repr}}

shell.load_config(
	bcftools = bcftools,
	vcftools = vcftools,
	gatk     = gatk,
)

para = Parallel(nthread, raiseExc = True)
invcfs = para.run(vcfIndex, [
	(vcf, tabix) for vcf in invcfs
])
# don't reuse it
del para


def run_vcftools():
	params.d       = params.get('d', True)
	params.t       = params.get('t', True)
	params.R       = params.get('R', '0/0')
	params._       = invcfs
	params._stdout = outfile[:-3] if gz else outfile
	shell.vcftools(**params)
	if gz:
		shell.bgzip(outfile[:-3])

def run_bcftools():
	params.F       = params.get('F', '+')
	params.o       = outfile
	params.threads = nthread
	params._       = invcfs
	params['0']    = params.get('0', True)
	if gz:
		params.O = 'z'
	shell.bcftools.merge(**params)

def run_gatk():
	params.T                   = 'CombineVariants'
	params.o                   = outfile
	params.R                   = ref
	params.nt                  = nthread
	params.variant             = invcfs
	params._stdout             = outfile[:-3] if gz else outfile
	params.genotypemergeoption = params.get('genotypemergeoption', 'UNIQUIFY')
	shell.gatk(**params)
	if gz:
		shell.bgzip(outfile[:-3])

tools = dict(
	vcftools = run_vcftools,
	bcftools = run_bcftools,
	gatk     = run_gatk,
)

try:
	tools[tool]()
except KeyError:
	raise KeyError('Tool {!r} not supported.'.format(tool))
except:
	raise
finally:
	pass


from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell, logger
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
dtoxog  = {{args.dtoxog | quote}}
params  = {{args.params | repr}}
keep    = {{args.keep | repr}}
nthread = {{args.nthread | repr}}

shell.load_config(dtoxog = dtoxog)

# see if maf file has required fields
# required fields
# Chromosome -- Contig number without a prefix.  E.g. "3" or "X" (without quotes)
# Start_position -- Position of the SNV
# End_position -- Should be the same as Start_position for SNV
# Reference_Allele -- The allele found in the reference genome at this position.  Single character representing the base, e.g. "G".
# Tumor_Seq_Allele1 -- alternate allele.  Single character representing the base, e.g. "G".
# Tumor_Sample_Barcode -- name of the tumor (or "case") sample.  This is used to generate file names and plots.
# Matched_Norm_Sample_Barcode -- name of the normal (or "control") sample.  This is used to generate file names and plots.
# ref_context -- Small window into the reference at the SNV.  The center position should be the same as Reference_Allele.  The total string should be of odd length and have a minimum length of 3.
# 	For example: Reference Allele is G, Chromosome is 1, Start_position and End_position are 120906037:  ref_context is CTTTTTTCGCGCAAAAATGCC  (string size is 21, in this case)
# i_t_ALT_F1R2 -- the number of reads with pair orientation of F1R2 and with the alternate allele (Tumor_Seq_Allele1).
# i_t_ALT_F2R1 -- the number of reads with pair orientation of F2R1 and with the alternate allele (Tumor_Seq_Allele1).
# i_t_REF_F1R2 -- the number of reads with pair orientation of F1R2 and with the reference allele (Reference_Allele).
# i_t_REF_F2R1 -- the number of reads with pair orientation of F2R1 and with the reference allele (Reference_Allele).
# i_t_Foxog -- Foxog, as described in the methods.  Depending on the nature of the reference and alternate alleles, either i_t_ALT_F1R2/(i_t_ALT_F1R2 + i_t_ALT_F2R1)  or i_t_ALT_F2R1/(i_t_ALT_F1R2 + i_t_ALT_F2R1).
# 	C>anything:  numerator is i_t_ALT_F2R1
# 	A>anything:  numerator is i_t_ALT_F2R1
# 	G>anything:  numerator is i_t_ALT_F1R2
# 	T>anything:  numerator is i_t_ALT_F1R2
# Variant_Type -- "SNP" (without quotes)
# i_picard_oxoQ -- will be 0

required_fields = { # set
	'Chromosome',
	'Start_Position',
	'End_Position',
	'Reference_Allele',
	'Tumor_Seq_Allele1',
	'Tumor_Sample_Barcode',
	'Matched_Norm_Sample_Barcode',
	'ref_context',
	'i_t_ALT_F1R2',
	'i_t_ALT_F2R1',
	'i_t_REF_F1R2',
	'i_t_REF_F2R1',
	'i_t_Foxog',
	'Variant_Type',
	'i_picard_oxoQ',
}

# check
logger.info('Checking required fields ...')
header = None
with open(infile) as f:
	for line in f:
		if line.startswith('Hugo_Symbol'):
			header = line.rstrip('\n').split('\t')
			break

if not header:
	raise ValueError('No header found for input MAF file.')

fields_not_found = required_fields - set(header)
if fields_not_found and fields_not_found != {'i_picard_oxoQ'}:
	raise ValueError('Required fields not found: %s' % (fields_not_found))


logger.info('Cleaning up input MAF file ...')
# Add i_picard_oxoQ as 0 if column does not exist
# TODO: add real i_picard_oxoQ
# Set i_t_ALT_F1R2, i_t_ALT_F2R1, i_t_REF_F1R2, i_t_REF_F2R1 as 0 if they are not reported
# Exclude mutations other than [SNP, DNP, TNP]
infile_fixed = path.join(path.dirname(outfile), 'fixed.' + path.basename(infile))
reader = TsvReader(infile, comment = '#', cnames = True)
writer = TsvWriter(infile_fixed)
if fields_not_found:
	writer.cnames = reader.cnames + ['i_picard_oxoQ']
writer.writeHead(lambda cnames: [
	'Start_position' if cname == 'Start_Position' else
	'End_position' if cname == 'End_Position' else cname
	for cname in cnames])
for r in reader:
	if r.Variant_Type not in ('SNP', 'DNP', 'TNP'):
		continue
	r.i_picard_oxoQ = 0
	for key in ('i_t_ALT_F1R2', 'i_t_ALT_F2R1', 'i_t_REF_F1R2', 'i_t_REF_F2R1'):
		if not r[key].isdigit():
			r[key] = 0
	writer.write(r)
writer.close()

params.mafFilename       = infile_fixed
params.outputMAFFilename = path.basename(outfile) + '.all'
params.outputDir         = path.dirname(outfile)
params.nthread           = nthread

# avoid lost values
for key, value in params.items():
	if isinstance(value, bool):
		params[key] = int(value)

shell.mkdir(path.join(path.dirname(outfile), 'figures'))
logger.info('Running DToxoG ...')
shell.fg.dtoxog(**params)


logger.info('Fixing output file format issues ...')
if not keep:
	# remove variants with isArtifactMode = 1
	with open(outfile + '.all') as fin, open(outfile, 'w') as fout:
		for line in fin:
			if line.startswith('#'):
				fout.write(line)
			else:
				break
	reader = TsvReader(outfile + '.all')
	writer = TsvWriter(outfile, append = True)
	writer.cnames = reader.cnames
	writer.writeHead(lambda cnames: [
		'Start_Position' if cname == 'Start_position' else
		'End_Position' if cname == 'End_position' else cname
		for cname in cnames])
	for r in reader:
		if r.isArtifactMode == '1':
			continue
		writer.write(r)
	reader.close()
	writer.close()
else:
	# get Start_Position and End_Position back
	with open(outfile + '.all') as fin, open(outfile, 'w') as fout:
		for line in fin:
			if line.startswith('Hugo_Symbol'):
				fout.write(line.replace('Start_position', 'Start_Position').replace('End_position', 'End_Position'))
			else:
				fout.write(line)



from pyppl import Box
from bioprocs.utils import shell2 as shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
bcftools = {{args.bcftools | quote}}
params   = {{args.params | repr}}

shell.load_config(bcftools = bcftools)

params._ = infile
params._out = outfile
shell.bcftools.query(**params)



from os import path
from pyppl import Box
from bioprocs.utils import shell

infile    = {{ i.infile | quote}}
outfile   = {{ o.outfile | quote }}
joboutdir = {{job.outdir | quote}}
tool      = {{ args.tool | quote}}
bedtools  = {{ args.bedtools | quote}}
bedops    = {{ args.bedops | quote}}
argsmem   = {{ args.mem | quote}}
sortby    = {{ args.by | quote}}
unique    = {{ args.unique | bool}}
params    = {{ args.params | repr}}
tmpdir    = {{ args.tmpdir | quote}}
chrorder  = {{ args.chrorder if not args.chrorder or isinstance(args.chrorder, (tuple, list)) else args.chrorder.split(',') | repr }}

shell.TOOLS.bedtools = bedtools
shell.TOOLS.bedops   = bedtools

def run_sort():
	params.T = tmpdir
	params.S = argsmem
	params.u = unique
	if sortby == 'coord':
		params.k = ['1,1', '2,2n']
	else:
		params.k = '4'
	shell.grep('^#', infile, _stdout = outfile)
	shell.Shell().grep(v = '^#', _ = infile).pipe(dash = '-', equal = ' ', duplistkey = True).sort(__stdout = outfile, **params).run()

def run_bedops():
	params['max-mem'] = args.mem
	params.tmpdir = tmpdir
	params._ = infile
	shell.grep('^#', infile, _stdout = outfile)
	c = shell.Shell(equal = ' ')
	if unique:
		c.bedops(**params).pipe().uniq(__stdout = outfile).run()
	else:
		params.__stdout = outfile
		c.bedops(**params).run()
	
def run_bedtools():
	params.i = infile
	shell.grep('^#', infile, _stdout = outfile)
	c = shell.Shell(dash = '-', equal = ' ', subcmd = True).bedtools
	if unique:
		c.sort(**params).pipe().uniq(__stdout = outfile).run()
	else:
		params.__stdout = outfile
		c.sort(**params).run()

def attach_order():
	if tool != 'sort':
		return
	mappings = dict(zip(chrorder, sorted(chrorder)))
	global infile
	tmpfile = path.join(joboutdir, path.basename(infile) + '.withorder')
	from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
	reader = TsvReader(infile, cnames = False)
	writer = TsvWriter(tmpfile)
	for r in reader:
		r[0] = mappings.get(r[0], r[0])
		writer.write(r)
	reader.close()
	writer.close()
	infile = tmpfile
	return dict(zip(mappings.values(), mappings.keys()))

def detach_order(mappings):
	if tool != 'sort' or not mappings:
		return
	shell.rm(infile)
	tmpfile = path.join(joboutdir, path.basename(infile) + '.withoutorder')
	from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
	reader = TsvReader(outfile, cnames = False)
	writer = TsvWriter(tmpfile)
	for r in reader:
		r[0] = mappings.get(r[0], r[0])
		writer.write(r)
	reader.close()
	writer.close()
	shell.mv(tmpfile, outfile)

tools = dict(
	sort     = run_sort,
	bedops   = run_bedops,
	bedtools = run_bedtools
)

try:
	if chrorder:
		mappings = attach_order()
	tools[tool]()
	if chrorder:
		detach_order(mappings)
except KeyError:
	raise KeyError('Tool {!r} not supported.'.format(tool))
except:
	raise
finally:
	pass


from bioprocs.utils import shell2 as shell
from bioprocs.utils.tsvio2 import TsvWriter
from gff import Gff

infile  = {{i.infile | quote}}
bedfile = {{i.bedfile | quote}}
bedext  = {{i.bedfile | ?:ext(_) == '.gz' | prefix | :_ | ext | [1:] | quote }}
outfile = {{o.outfile | quote}}
bwtool  = {{args.bwtool | quote}}

if 'bed' not in bedext:
	tmpfile = path.join(path.dirname(outfile), 'regions.bed')
	gff = Gff(bedfile)
	writer = TsvWriter(tmpfile)
	for r in gff:
		writer.write([
			r['seqid'], r['start'], r['end'],
			r['attributes'].get(
				'id',
				r['attributes'].get('name', f'{r["seqid"]}:{r["start"]}-{r["end"]}'))
		])
	bedfile = tmpfile

shell.load_config(bwtool = dict(_exe = bwtool, _prefix = '-'))

shell.fg.bwtool.extract('bed', bedfile, infile, outfile, tabs = True)



"""Processes for BED files"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

from .bedtools import *

@procfactory
def _pBedSort():
	"""
	@name:
		pBedSort
	@description:
		Sort bed files
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`tool`:     The tool used to sort the file. Default: sort (bedtools, bedops)
		`bedtools`: The path to bedtools. Default: bedtools
		`bedops`:   The path to bedops' sort-bed. Default: sort-bed
		`mem`:      The memory to use. Default: 8G
		`by`:       Sort by coordinates("coord", default) or name("name")
			- Only available when use tool `sort`
		`tmpdir`:   The tmpdir to use. Default: `$TMPDIR`
		`unique`:   Remove the dupliated records? Default: True
		`params`:   Other params for `tool`. Default: {}
		`chrorder`: The chromosome order used to sort. Default: `None`
			- `None`: Sort by natural order (chr1 followed by chr10, instead of chr2)
			- Only available when using `sort` (`args.tool = 'sort'`)
	@requires:
		[`bedtools`](http://bedtools.readthedocs.io/en/latest/index.html)
		[`bedops`](https://github.com/bedops/bedops)
	"""
	pBedSort               = Proc(desc = 'Sort bed files.')
	pBedSort.input         = "infile:file"
	pBedSort.output        = "outfile:file:{{i.infile | bn}}"
	pBedSort.args.tool     = 'sort'
	pBedSort.args.bedtools = params.bedtools.value
	pBedSort.args.bedops   = params.bedops_sort.value
	pBedSort.args.mem      = '8G'
	pBedSort.args.by       = 'coord'
	pBedSort.args.unique   = True
	pBedSort.args.params   = Box()
	pBedSort.args.chrorder = None
	pBedSort.args.tmpdir   = params.tmpdir.value
	pBedSort.lang          = params.python.value
	pBedSort.script        = "file:scripts/bed/pBedSort.py"
	return pBedSort

@procfactory
def _pBedLiftover():
	"""
	@name:
		pBedLiftover
	@description:
		Lift over bed files.
	@input:
		`infile:file`: The input bed file
	@output:
		`outfile:file`: The output file
		`umfile:file` : The unmapped file
	@args:
		`liftover`: The liftover program
		`lochain` : the liftover chain file
	@require:
		`liftover` from UCSC
	"""
	pBedLiftover               = Proc(desc = 'Lift over bed files.')
	pBedLiftover.input         = 'infile:file'
	pBedLiftover.output        = 'outfile:file:{{i.infile | bn}}, umfile:file:{{i.infile | fn}}.unmapped{{i.infile | ext}}'
	pBedLiftover.args.liftover = params.liftover.value
	pBedLiftover.args.lochain  = params.lochain.value
	pBedLiftover.args.params   = Box()
	pBedLiftover.lang          = params.python.value
	pBedLiftover.script        = "file:scripts/bed/pBedLiftover.py"
	return pBedLiftover

@procfactory
def _pGff2Bed(alias = 'pBedFromGff'):
	"""
	@input:
		infile: The input gtf/gff file
	@output:
		outfile: The converted bed file
	@args:
		bedcols:  Strings of python functions used to convert GTF/GFF records to BED fields.
			- You can define the NAME column here, and extra columns after the 6th column.
			- For example: `args.bedcols = {"NAME": "lambda attrs: rec.CHR + ':' + rec.START"}`
				- `attrs` are the attributes of GFF records, plus CHR, START, END, SCORE and STRAND.
				- See: https://github.com/pwwang/pygff
				- By default, NAME will use `id` in attributes, and then `name`. Otherwise `CHR:START-END` will  be used.
			- You can also add extra columns starting from 7th column of BED file, for example:
				- `args.bedcols = {"CN": "lambda attrs: attrs['CopyNumber']"}`
		keepattrs: Keep the original attributes at last column of the output BED file.
		outhead: Put head to output file or not.
			- Could be prefix to the head.
	"""
	return Box(
		desc   = 'Convert GTF/GFF file to BED file',
		lang   = params.python.value,
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | stem}}.bed',
		args   = Box(bedcols = Box(), keepattrs = True, outhead = '#')
	)
